from __future__ import annotations

import asyncio
import inspect
import time
import uuid
from collections import defaultdict
from collections.abc import Awaitable, Callable, Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, cast

from fastapi import Request
from fastapi.responses import Response
from fastapi.templating import Jinja2Templates
from typing_extensions import Self

from . import background_tasks, binding, core, helpers, json, storage
from .awaitable_response import AwaitableResponse
from .dependencies import generate_resources
from .element import Element
from .favicon import get_favicon_url
from .javascript_request import JavaScriptRequest
from .logging import log
from .observables import ObservableDict
from .outbox import Outbox
from .sub_pages_router import SubPagesRouter
from .translations import translations
from .version import __version__

if TYPE_CHECKING:
    from .page import page

templates = Jinja2Templates(Path(__file__).parent / 'templates')

HTML_ESCAPE_TABLE = str.maketrans({
    '&': '&amp;',
    '<': '&lt;',
    '>': '&gt;',
    '`': '&#96;',
    '$': '&#36;',
})

HEADWIND_CONTENT = (Path(__file__).parent / 'static' / 'headwind.css').read_text().strip()


class ClientConnectionTimeout(TimeoutError):
    def __init__(self, client: Client) -> None:
        super().__init__(f'ClientConnectionTimeout: {client.id}')
        self.client = client


class Client:
    page_routes: ClassVar[dict[Callable[..., Any], str]] = {}
    '''Maps page builders to their routes.'''

    instances: ClassVar[dict[str, Client]] = {}
    '''Maps client IDs to clients.'''

    shared_head_html = ''
    '''HTML to be inserted in the <head> of every page template.'''

    shared_body_html = ''
    '''HTML to be inserted in the <body> of every page template.'''

    def __init__(self, page: page, *, request: Request | None = None) -> None:
        self._request = request
        self.id = str(uuid.uuid4())
        self.created = time.time()
        self.instances[self.id] = self

        self.elements: dict[int, Element] = {}
        self.next_element_id: int = 0
        self._waiting_for_connection = asyncio.Event()
        self._waiting_for_disconnect = asyncio.Event()
        self._connected = asyncio.Event()
        self._deleted_event = asyncio.Event()
        self.environ: dict[str, Any] | None = None
        self.on_air = False
        self._num_connections: defaultdict[str, int] = defaultdict(int)
        self._delete_tasks: dict[str, asyncio.Task] = {}
        self._deleted = False
        self._socket_to_document_id: dict[str, str] = {}
        self.tab_id: str | None = None
        self._exception_handlers: list[Callable[[Exception], Any] | Callable[[], Any]] = []

        self.page = page
        self.outbox = Outbox(self)

        if self._request is not None:
            self._request.scope['nicegui_page_path'] = self.page.path

        with Element('q-layout', _client=self).props('view="hhh lpr fff"').classes('nicegui-layout') as self.layout:
            with Element('q-page-container') as self.page_container:
                with Element('q-page'):
                    self.content = Element('div').classes('nicegui-content')

        self.title: str | None = None

        self._head_html = ''
        self._body_html = ''

        self.storage = ObservableDict()

        self.connect_handlers: list[Callable[..., Any] | Awaitable] = []
        self.disconnect_handlers: list[Callable[..., Any] | Awaitable] = []
        self.delete_handlers: list[Callable[..., Any] | Awaitable] = []

        self._temporary_socket_id: str | None = None

        with self:
            self.sub_pages_router = SubPagesRouter(request)

    @property
    def request(self) -> Request:
        """The request object for the client."""
        if self._request is None:
            raise RuntimeError('Request is not set')
        return self._request

    @property
    def ip(self) -> str:
        """The IP address of the client.

        *Updated in version 2.0.0: The IP address is available even before the client connects.*
        *Updated in version 3.0.0: The IP address is always defined (never ``None``).*
        """
        return self.request.client.host if self.request.client is not None else ''

    @property
    def has_socket_connection(self) -> bool:
        """Whether the client is connected."""
        return self.tab_id is not None

    @property
    def head_html(self) -> str:
        """The HTML code to be inserted in the <head> of the page template."""
        return self.shared_head_html + self._head_html

    @property
    def body_html(self) -> str:
        """The HTML code to be inserted in the <body> of the page template."""
        return self.shared_body_html + self._body_html

    def __enter__(self) -> Self:
        self.content.__enter__()
        return self

    def __exit__(self, *_) -> None:
        self.content.__exit__()

    def build_response(self, request: Request, status_code: int = 200) -> Response:
        """Build a FastAPI response for the client."""
        self.outbox.updates.clear()
        prefix = request.headers.get('X-Forwarded-Prefix', '') + request.scope.get('root_path', '')
        elements = json.dumps({
            id: element._to_dict() for id, element in self.elements.items()  # pylint: disable=protected-access
        })
        socket_io_js_query_params = {
            **core.app.config.socket_io_js_query_params,
            'client_id': self.id,
            'next_message_id': self.outbox.next_message_id,
            'implicit_handshake': not _is_prefetch(request),
        }
        vue_html, vue_styles, vue_scripts, imports, js_imports, js_imports_urls = \
            generate_resources(prefix, self.elements.values())
        return templates.TemplateResponse(
            request=request,
            name='index.html',
            context={
                'request': request,
                'version': __version__,
                'elements': elements.translate(HTML_ESCAPE_TABLE),
                'head_html': self.head_html,
                'body_html': '<style>' + '\n'.join(vue_styles) + '</style>\n' + self.body_html + '\n' + '\n'.join(vue_html),
                'vue_scripts': '\n'.join(vue_scripts),
                'imports': json.dumps(imports),
                'js_imports': '\n'.join(js_imports),
                'js_imports_urls': js_imports_urls,
                'vue_config': json.dumps(core.app.config.quasar_config),
                'vue_config_script': core.app.config.vue_config_script,
                'title': self.resolve_title(),
                'viewport': self.page.resolve_viewport(),
                'favicon_url': get_favicon_url(self.page, prefix),
                'dark': str(self.page.resolve_dark()),
                'language': self.page.resolve_language(),
                'translations': translations.get(self.page.resolve_language(), translations['en-US']),
                'prefix': prefix,
                'tailwind': core.app.config.tailwind,
                'unocss': core.app.config.unocss,
                'headwind_css': HEADWIND_CONTENT if core.app.config.tailwind else '',
                'prod_js': core.app.config.prod_js,
                'socket_io_js_query_params': socket_io_js_query_params,
                'socket_io_js_extra_headers': core.app.config.socket_io_js_extra_headers,
                'socket_io_js_transports': core.app.config.socket_io_js_transports,
            },
            status_code=status_code,
            headers={'Cache-Control': 'no-store', 'X-NiceGUI-Content': 'page'},
        )

    def resolve_title(self) -> str:
        """Return the title of the page."""
        return self.page.resolve_title() if self.title is None else self.title

    async def connected(self, timeout: float | None = None) -> None:
        """Block execution until the client is connected.

        :param timeout: timeout in seconds (default: ``None``)
        """
        if self.has_socket_connection:
            return
        self._waiting_for_connection.set()
        self._connected.clear()
        try:
            await asyncio.wait_for(self._connected.wait(), timeout=None if _is_prefetch(self.request) else timeout)
        except asyncio.TimeoutError as e:
            raise ClientConnectionTimeout(self) from e

    async def disconnected(self) -> None:
        """Block execution until the client disconnects."""
        if not self.has_socket_connection:
            await self.connected()
        if self.id in self.instances:
            self._waiting_for_disconnect.set()
            self._deleted_event.clear()
            await self._deleted_event.wait()

    def run_javascript(self, code: str, *, timeout: float = 1.0) -> AwaitableResponse:
        """Execute JavaScript on the client.

        If the function is awaited, the result of the JavaScript code is returned.
        Otherwise, the JavaScript code is executed without waiting for a response.

        Obviously the JavaScript code is only executed after the client is connected.
        Internally, ``await client.connected()`` is called before the JavaScript code is executed (*since version 3.0.0*).
        This might delay the execution of the JavaScript code and is not covered by the ``timeout`` parameter.

        :param code: JavaScript code to run
        :param timeout: timeout in seconds (default: 1.0)

        :return: AwaitableResponse that can be awaited to get the result of the JavaScript code
        """
        request_id = str(uuid.uuid4())
        target_id = self._temporary_socket_id or self.id

        def send_and_forget():
            self.outbox.enqueue_message('run_javascript', {'code': code}, target_id)

        async def send_and_wait():
            self.outbox.enqueue_message('run_javascript', {'code': code, 'request_id': request_id}, target_id)
            await self.connected()
            return await JavaScriptRequest(request_id, timeout=timeout)

        return AwaitableResponse(send_and_forget, send_and_wait)

    def open(self, target: Callable[..., Any] | str, new_tab: bool = False) -> None:
        """Open a new page in the client."""
        path = target if isinstance(target, str) else self.page_routes[target]
        self.outbox.enqueue_message('open', {'path': path, 'new_tab': new_tab}, self.id)

    def download(self, src: str | bytes, filename: str | None = None, media_type: str = '') -> None:
        """Download a file from a given URL or raw bytes."""
        self.outbox.enqueue_message('download', {'src': src, 'filename': filename, 'media_type': media_type}, self.id)

    def on_connect(self, handler: Callable[..., Any] | Awaitable) -> None:
        """Add a callback to be invoked when the client connects.

        The callback has an optional parameter of `nicegui.Client`.
        """
        self.connect_handlers.append(handler)

    def on_disconnect(self, handler: Callable[..., Any] | Awaitable) -> None:
        """Add a callback to be invoked when the client disconnects.

        The callback has an optional parameter of `nicegui.Client`.

        *Updated in version 3.0.0: The handler is also called when a client reconnects.*
        """
        self.disconnect_handlers.append(handler)

    def on_delete(self, handler: Callable[..., Any] | Awaitable) -> None:
        """Add a callback to be invoked when the client is deleted.

        The callback has an optional parameter of `nicegui.Client`.

        *Added in version 3.0.0*
        """
        self.delete_handlers.append(handler)

    def on_exception(self, handler: Callable[[Exception], Any] | Callable[[], Any]) -> None:
        """Add a callback to be invoked for in-page exceptions (after the page has been sent to the browser).

        The callback has an optional parameter of `Exception`.
        """
        self._exception_handlers.append(handler)

    def handle_handshake(self, socket_id: str, document_id: str, next_message_id: int | None) -> None:
        """Cancel pending disconnect task and invoke connect handlers. (For internal use only.)"""
        self._waiting_for_connection.clear()
        self._connected.set()
        self._socket_to_document_id[socket_id] = document_id
        self._cancel_delete_task(document_id)
        self._num_connections[document_id] += 1
        if next_message_id is not None:
            self.outbox.try_rewind(next_message_id)
        storage.request_contextvar.set(self.request)
        for t in self.connect_handlers:
            self.safe_invoke(t)
        for t in core.app._connect_handlers:  # pylint: disable=protected-access
            self.safe_invoke(t)

    def handle_disconnect(self, socket_id: str) -> None:
        """Wait for the browser to reconnect; invoke deletion handlers if it doesn't. (For internal use only.)"""
        if socket_id not in self._socket_to_document_id:
            return
        document_id = self._socket_to_document_id.pop(socket_id)
        self._cancel_delete_task(document_id)
        self._num_connections[document_id] -= 1
        tab_id_to_close = self.tab_id
        self.tab_id = None

        for t in self.disconnect_handlers:
            self.safe_invoke(t)
        for t in core.app._disconnect_handlers:  # pylint: disable=protected-access
            self.safe_invoke(t)

        async def delete_content() -> None:
            await asyncio.sleep(self.page.resolve_reconnect_timeout())
            if self._num_connections[document_id] == 0:
                self._num_connections.pop(document_id)
                self._delete_tasks.pop(document_id)
                await core.app.storage.close_tab(tab_id_to_close)
                self.delete()
        self._delete_tasks[document_id] = \
            background_tasks.create(delete_content(), name=f'delete content {document_id}')

    def _cancel_delete_task(self, document_id: str) -> None:
        if document_id in self._delete_tasks:
            self._delete_tasks.pop(document_id).cancel()

    def handle_event(self, msg: dict) -> None:
        """Forward an event to the corresponding element. (For internal use only.)"""
        with self:
            sender = self.elements.get(msg['id'])
            if sender is not None and not sender.is_ignoring_events:
                msg['args'] = [None if arg is None else json.loads(arg) for arg in msg.get('args', [])]
                if len(msg['args']) == 1:
                    msg['args'] = msg['args'][0]
                sender._handle_event(msg)  # pylint: disable=protected-access

    def handle_log_message(self, msg: dict) -> None:
        """Log a message from the client. (For internal use only.)"""
        {
            'debug': log.debug,
            'info': log.info,
            'warning': log.warning,
            'error': log.error,
        }[msg['level']](msg['message'])

    def handle_javascript_response(self, msg: dict) -> None:
        """Store the result of a JavaScript command. (For internal use only.)"""
        JavaScriptRequest.resolve(msg['request_id'], msg.get('result'))

    def safe_invoke(self, func: Callable[..., Any] | Awaitable) -> None:
        """Invoke the potentially async function in the client context and catch any exceptions."""
        func_name = func.__name__ if hasattr(func, '__name__') else str(func)
        try:
            if isinstance(func, Awaitable):
                async def func_with_client():
                    with self:
                        await func
                background_tasks.create(func_with_client(), name=f'func with client {self.id} {func_name}')
            else:
                with self:
                    result = func(self) if len(inspect.signature(func).parameters) == 1 else func()
                if helpers.is_coroutine_function(func) and not isinstance(result, asyncio.Task):
                    async def result_with_client():
                        with self:
                            await result
                    background_tasks.create(result_with_client(), name=f'result with client {self.id} {func_name}')
        except Exception as e:
            core.app.handle_exception(e)

    def remove_elements(self, elements: Iterable[Element]) -> None:
        """Remove the given elements from the client."""
        element_list = list(elements)  # NOTE: we need to iterate over the elements multiple times
        binding.remove(element_list)
        for element in element_list:
            element._handle_delete()  # pylint: disable=protected-access
            element._deleted = True  # pylint: disable=protected-access
            self.outbox.enqueue_delete(element)
            self.elements.pop(element.id, None)

    def remove_all_elements(self) -> None:
        """Remove all elements from the client."""
        self.remove_elements(self.elements.values())

    def handle_exception(self, exception: Exception) -> None:
        """Handle an in-page exception by invoking handlers registered via `ui.on_exception(...)`."""
        for handler in self._exception_handlers:
            with self.content:
                if helpers.expects_arguments(handler):
                    result = cast(Callable[[Exception], Any], handler)(exception)
                else:
                    result = cast(Callable[[], Any], handler)()
            if helpers.is_coroutine_function(handler):
                async def wait_for_result(result: Any = result) -> None:
                    with self.content:
                        await result
                background_tasks.create(wait_for_result(), name=f'UI exception {handler.__name__}')

    def delete(self) -> None:
        """Delete a client and all its elements.

        If the global clients dictionary does not contain the client, its elements are still removed and a KeyError is raised.
        Normally this should never happen, but has been observed (see #1826).
        """
        for t in self.delete_handlers:
            self.safe_invoke(t)
        for t in core.app._delete_handlers:  # pylint: disable=protected-access
            self.safe_invoke(t)
        self._waiting_for_disconnect.clear()
        self._deleted_event.set()
        self.remove_all_elements()
        self.outbox.stop()
        del Client.instances[self.id]
        self._deleted = True
        self._connected.set()  # for terminating connected() waits
        self._connected.clear()

    def check_existence(self) -> None:
        """Check if the client still exists and print a warning if it doesn't."""
        if self._deleted:
            helpers.warn_once('Client has been deleted but is still being used. '
                              'This is most likely a bug in your application code. '
                              'See https://github.com/zauberzeug/nicegui/issues/3028 for more information.',
                              stack_info=True)

    @classmethod
    def prune_instances(cls, *, client_age_threshold: float = 60.0) -> None:
        """Prune stale clients."""
        try:
            stale_clients = [
                client
                for client in cls.instances.values()
                if (
                    not client.has_socket_connection and
                    not client._delete_tasks and  # pylint: disable=protected-access
                    client.created <= time.time() - client_age_threshold
                )
            ]
            for client in stale_clients:
                log.debug(f'Pruning stale client {client.id}')
                client.delete()

        except Exception:
            log.exception('Error while pruning clients')


def _is_prefetch(request: Request) -> bool:
    purpose = (request.headers.get('Sec-Purpose') or request.headers.get('Purpose') or '').lower()
    return 'prefetch' in purpose and 'prerender' not in purpose



import asyncio
import inspect
import re
from collections.abc import Callable
from typing import Any
from urllib.parse import urlparse

from starlette.datastructures import QueryParams
from typing_extensions import Self

from .. import background_tasks, json
from ..context import context
from ..element import Element
from ..elements.label import Label
from ..functions.javascript import run_javascript
from ..logging import log
from ..page_arguments import PageArguments, RouteMatch


class SubPages(Element, component='sub_pages.js', default_classes='nicegui-sub-pages'):

    def __init__(self,
                 routes: dict[str, Callable] | None = None,
                 *,
                 root_path: str | None = None,
                 data: dict[str, Any] | None = None,
                 show_404: bool = True,
                 ) -> None:
        """Create a container for client-side routing within a page.

        Provides URL-based navigation between different views to build single page applications (SPAs).
        Routes are defined as path patterns mapping to page builder functions.
        Path parameters like "/user/{id}" are extracted and passed to the builder function.

        *Added in version 2.22.0*

        :param routes: dictionary mapping path patterns to page builder functions
        :param root_path: path prefix to strip from incoming paths (ignored by nested ``ui.sub_pages`` elements)
        :param data: arbitrary data passed to all page builder functions
        :param show_404: whether to show a 404 error message if the full path could not be consumed
            (can be useful for dynamically created nested sub pages) (default: ``True``)
        """
        super().__init__()
        self._router = context.client.sub_pages_router
        self._routes = routes or {}
        parent_sub_pages_element = next((el for el in self.ancestors() if isinstance(el, SubPages)), None)
        self._rendered_path = ''
        self._root_path = parent_sub_pages_element._rendered_path if parent_sub_pages_element else root_path
        self._data = data or {}
        self._match: RouteMatch | None = None
        self._active_tasks: set[asyncio.Task] = set()
        self._404_enabled = show_404
        self.has_404 = False
        self._show()

    def add(self, path: str, page: Callable) -> Self:
        """Add a new route.

        :param path: path pattern to match (e.g., "/user/{id}" for parameterized routes)
        :param page: function to call when this path is accessed
        :return: self for method chaining
        """
        self._routes[path] = page
        self._show()
        return self

    def refresh(self) -> None:
        """Rebuild this sub pages element.

        *Added in version 3.1.0*
        """
        self._reset_match()
        self._show()

    def _show(self) -> None:
        """Display the page matching the current URL path."""
        self._rendered_path = ''
        match = self._find_matching_path()
        # NOTE: if path and query params are the same, only update fragment without re-rendering
        if (
            match is not None and
            self._match is not None and
            match.path == self._match.path and
            not self._required_query_params_changed(match) and
            not (self.has_404 and self._match.remaining_path == match.remaining_path)
        ):
            # NOTE: Even though our matched path is the same, the remaining path might still require us to handle 404 (if we are the last sub pages element)
            if match.remaining_path and not any(isinstance(el, SubPages) for el in self.descendants()):
                self._set_match(None)
            else:
                self._handle_scrolling(match, behavior='smooth')
                self._set_match(match)
        else:
            self._cancel_active_tasks()
            with self.clear():
                if match is not None and self._render_page(match):
                    self._set_match(match)
                else:
                    self._set_match(None)

    def _render_page(self, match: RouteMatch) -> bool:
        kwargs = PageArguments.build_kwargs(match, self, self._data)
        self._rendered_path = f'{self._root_path or ""}{match.path}'
        try:
            result = match.builder(**kwargs)
        except Exception as e:
            self.clear()  # NOTE: clear partial content created before the exception
            self._render_error(e)
            self.client.handle_exception(e)
            return True

        self._handle_scrolling(match, behavior='instant')
        if asyncio.iscoroutine(result):
            async def background_task():
                with self:
                    try:
                        await result
                    except Exception as e:
                        self.client.handle_exception(e)
                        raise

            task = background_tasks.create(background_task(), name=f'building sub_page {match.pattern}')
            self._active_tasks.add(task)

            def _close_if_canceled(t: asyncio.Task) -> None:
                if t.cancelled():
                    result.close()
                self._active_tasks.discard(t)

            task.add_done_callback(_close_if_canceled)
        return True

    def _render_404(self) -> None:
        """Display a 404 error message for unmatched routes."""
        Label(f'404: sub page {self._router.current_path} not found')

    def _render_error(self, _: Exception) -> None:  # NOTE: exception is exposed for debugging scenarios via inheritance
        msg = f'sub page {self._router.current_path} produced an error'
        Label(f'500: {msg}')
        log.error(msg, exc_info=True)

    def _set_match(self, match: RouteMatch | None) -> None:
        self._match = match
        self.has_404 = match is None
        if self.has_404 and self._404_enabled:
            with self.clear():
                self._render_404()

    def _reset_match(self) -> None:
        self._match = None

    def _find_matching_path(self) -> RouteMatch | None:
        match: RouteMatch | None = None
        relative_path = self._router.current_path[len(self._root_path or ''):]
        if not relative_path.startswith('/'):
            relative_path = '/' + relative_path
        segments = relative_path.split('/')
        query_params: QueryParams | None = None
        while segments:
            path = '/'.join(segments)
            if not path:
                path = '/'
            match, query_params = self._match_route(path, query_params)
            if match is not None:
                match.remaining_path = urlparse(relative_path).path.rstrip('/')[len(match.path):]
                break
            segments.pop()
        return match

    def _match_route(self, path: str, query_params: QueryParams | None) -> tuple[RouteMatch | None, QueryParams | None]:
        parsed_url = urlparse(path)
        path_only = parsed_url.path.rstrip('/')
        query_params = query_params or QueryParams(parsed_url.query)
        fragment = parsed_url.fragment
        if not path_only.startswith('/'):
            path_only = '/' + path_only

        for route, builder in self._routes.items():
            parameters = self._match_path(route, path_only)
            if parameters is not None:
                return RouteMatch(
                    path=path_only,
                    pattern=route,
                    builder=builder,
                    parameters=parameters,
                    query_params=query_params,
                    fragment=fragment,
                ), query_params
        return None, query_params

    @staticmethod
    def _match_path(pattern: str, path: str) -> dict[str, str] | None:
        if '{' not in pattern:
            return {} if pattern == path else None

        regex_pattern = re.escape(pattern)
        for match in re.finditer(r'\\{(\w+)\\}', regex_pattern):
            param_name = match.group(1)
            regex_pattern = regex_pattern.replace(f'\\{{{param_name}\\}}', f'(?P<{param_name}>[^/]+)')

        regex_match = re.match(f'^{regex_pattern}$', path)
        return regex_match.groupdict() if regex_match else None

    def _cancel_active_tasks(self) -> None:
        for task in self._active_tasks:
            if not task.done() and not task.cancelled():
                task.cancel()
        self._active_tasks.clear()

    def _handle_scrolling(self, match: RouteMatch, *, behavior: str) -> None:
        if match.fragment:
            self._scroll_to_fragment(match.fragment, behavior=behavior)
        elif not self._router.is_initial_request:  # NOTE: the initial path has no fragment; to not interfere with later fragment scrolling, we skip scrolling to top
            self._scroll_to_top(behavior=behavior)

    def _scroll_to_fragment(self, fragment: str, *, behavior: str) -> None:
        run_javascript(f'''
            requestAnimationFrame(() => {{
                const frag = {json.dumps(fragment)};
                const el = document.getElementById(frag) || document.querySelector("a[name=" + JSON.stringify(frag) + "]");
                el?.scrollIntoView({{ behavior: "{behavior}" }});
            }});
        ''')

    def _scroll_to_top(self, *, behavior: str) -> None:
        run_javascript(f'''
            requestAnimationFrame(() => {{ window.scrollTo({{top: 0, left: 0, behavior: "{behavior}"}}); }});
        ''')

    def _required_query_params_changed(self, route_match: RouteMatch) -> bool:
        if self._match is None:
            return True
        current_params = route_match.query_params
        previous_params = self._match.query_params
        for name, param in inspect.signature(route_match.builder).parameters.items():
            if param.annotation is PageArguments:
                return current_params != previous_params
            if current_params.get(name) != previous_params.get(name):
                return True
        return False



from pathlib import Path

from typing_extensions import Self

from ...awaitable_response import AwaitableResponse
from ...defaults import DEFAULT_PROP, resolve_defaults
from ...element import Element
from ...events import GenericEventArguments, Handler, XtermBellEventArguments, XtermDataEventArguments, handle_event


class Xterm(Element, component='xterm.js', esm={'nicegui-xterm': 'dist'}):

    @resolve_defaults
    def __init__(self, options: dict | None = DEFAULT_PROP | None) -> None:
        """Xterm

        This element is a wrapper around `xterm.js <https://github.com/xtermjs/xterm.js>`_ to emulate a terminal.
        Note: This element provides only a front-end component without an underlying shell.

        *Added in version 3.1.0*

        :param options: A dictionary of options to configure the terminal, see the
                        `xterm.js documentation <https://xtermjs.org/docs/api/terminal/classes/terminal/#constructor>`_.
        """
        super().__init__()
        self.add_resource(Path(__file__).parent / 'dist')

        self._props['options'] = options or {}

    async def get_rows(self) -> int:
        """Get the number of rows in the terminal's viewport."""
        return await self.run_method('getRows')

    async def get_columns(self) -> int:
        """Get the number of columns  in the terminal's viewport."""
        return await self.run_method('getColumns')

    def fit(self) -> AwaitableResponse:
        """Fit the terminal to its container.

        This method only resizes the terminal, making the number of rows and columns match the size of the container.
        Note that you might also need to resize the backing pty to match the new size of the terminal.

        :return: AwaitableResponse that can be awaited to wait for the completion of the method call
        """
        return self.run_method('fit')

    def on_bell(self, callback: Handler[XtermBellEventArguments]) -> Self:
        """Add a callback to be invoked when the terminal's bell is triggered."""
        self.on('bell', lambda _: handle_event(callback, XtermBellEventArguments(sender=self, client=self.client)))
        return self

    def on_data(self, callback: Handler[XtermDataEventArguments]) -> Self:
        """Add a callback to be invoked when the user types or pastes into the terminal.

        In a typical setup, this should be passed on to the backing pty.
        """
        def handle_data(e: GenericEventArguments) -> None:
            handle_event(callback, XtermDataEventArguments(sender=self, client=self.client, data=e.args))
        self.on('data', handle_data)
        return self

    def input(self, data: str, *, was_user_input: bool = True) -> AwaitableResponse:
        """Input data to application side.

        The data is treated the same way input typed into the terminal would (i.e. the ``data`` event will fire).
        To write data onto the terminal (e.g. data coming from a pty), use the ``write`` method instead.

        :param data: The data to forward to the application.
        :param was_user_input: Whether the input is genuine user input.
                               This triggers additional behavior like focus or selection clearing.
                               Set this to ``False`` if the data sent should not be treated like user input would,
                               for example passing an escape sequence to the application.

        :return: AwaitableResponse that can be awaited to wait for the completion of the method call
        """
        return self.run_method('input', data, was_user_input)

    def write(self, data: bytes | str) -> AwaitableResponse:
        """Write data to the terminal.

        :param data: The data to write to the terminal.
                     This can either be UTF-8 encoded bytes or a string.

        :return: AwaitableResponse that can be awaited to wait for the completion of the method call
        """
        # Xterm.js accepts an `Uint8Array`, which we can get by converting `bytes` to a `list`
        return self.run_method('write', list(data) if isinstance(data, bytes) else data)

    def writeln(self, data: bytes | str) -> AwaitableResponse:
        """Write data to the terminal, followed by a break line character (``\\n``).

        :param data: The data to write to the terminal.
                     This can either be UTF-8 encoded bytes or a string.

        :return: AwaitableResponse that can be awaited to wait for the completion of the method call
        """
        # Xterm.js accepts an `Uint8Array`, which we can get by converting `bytes` to a `list`
        return self.run_method('writeln', list(data) if isinstance(data, bytes) else data)

    def run_terminal_method(self, name: str, *args, timeout: float = 1) -> AwaitableResponse:
        """Run a method of the Xterm.js terminal instance.

        Refer to the `Xterm.js documentation <https://xtermjs.org/docs/api/terminal/classes/terminal/#methods>`_
        for a list of methods.

        If the function is awaited, the result of the method call is returned.
        Otherwise, the method is executed without waiting for a response.

        :param name: name of the method (a prefix ":" indicates that the arguments are JavaScript expressions)
        :param args: arguments to pass to the method
        :param timeout: timeout in seconds (default: 1 second)

        :return: AwaitableResponse that can be awaited to get the result of the method call
        """
        return self.run_method('run_terminal_method', name, *args, timeout=timeout)



from ..element import Element


class Space(Element):

    def __init__(self) -> None:
        """Space

        This element is based on Quasar's `QSpace <https://quasar.dev/vue-components/space>`_ component.

        Its purpose is to simply fill all available space inside of a flexbox element.
        """
        super().__init__('q-space')



import asyncio
import inspect
import os
import platform
import signal
import urllib
from collections.abc import Awaitable, Callable, Iterator
from enum import Enum
from pathlib import Path
from typing import Any

from fastapi import FastAPI, HTTPException, Request, Response
from fastapi.responses import FileResponse

from .. import background_tasks, core, helpers
from ..client import Client
from ..context import context
from ..elements.mixins.color_elements import QUASAR_COLORS
from ..logging import log
from ..native import NativeConfig
from ..observables import ObservableSet
from ..server import Server
from ..staticfiles import CacheControlledStaticFiles
from ..storage import Storage
from .app_config import AppConfig
from .range_response import get_range_response


class State(Enum):
    STOPPED = 0
    STARTING = 1
    STARTED = 2
    STOPPING = 3


class App(FastAPI):
    from ..timer import Timer as timer  # pylint: disable=import-outside-toplevel

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs, docs_url=None, redoc_url=None, openapi_url=None)
        self.native = NativeConfig()
        self.storage = Storage()
        self.urls = ObservableSet()
        self._state: State = State.STOPPED
        self.config = AppConfig()

        self._startup_handlers: list[Callable[..., Any] | Awaitable] = []
        self._shutdown_handlers: list[Callable[..., Any] | Awaitable] = []
        self._connect_handlers: list[Callable[..., Any] | Awaitable] = []
        self._disconnect_handlers: list[Callable[..., Any] | Awaitable] = []
        self._delete_handlers: list[Callable[..., Any] | Awaitable] = []
        self._exception_handlers: list[Callable[..., Any]] = [log.exception]
        self._page_exception_handler: Callable[..., Any] | None = None

        self.colors()  # populate Quasar config with default colors

    @property
    def is_starting(self) -> bool:
        """Return whether NiceGUI is starting."""
        return self._state == State.STARTING

    @property
    def is_started(self) -> bool:
        """Return whether NiceGUI is started."""
        return self._state == State.STARTED

    @property
    def is_stopping(self) -> bool:
        """Return whether NiceGUI is stopping."""
        return self._state == State.STOPPING

    @property
    def is_stopped(self) -> bool:
        """Return whether NiceGUI is stopped."""
        return self._state == State.STOPPED

    def start(self) -> None:
        """Start NiceGUI. (For internal use only.)"""
        self._state = State.STARTING
        for t in self._startup_handlers:
            self.safe_invoke(t)
        self.on_shutdown(self.storage.on_shutdown)
        self.on_shutdown(background_tasks.teardown)
        self._state = State.STARTED

    async def stop(self) -> None:
        """Stop NiceGUI. (For internal use only.)"""
        self._state = State.STOPPING
        for t in self._shutdown_handlers:
            if isinstance(t, Awaitable):
                await t
            else:
                result = t(self) if len(inspect.signature(t).parameters) == 1 else t()
                if helpers.is_coroutine_function(t):
                    await result
        self._state = State.STOPPED

    def safe_invoke(self, func: Callable[..., Any] | Awaitable) -> None:
        """Invoke the potentially async function and catch any exceptions."""
        func_name = func.__name__ if hasattr(func, '__name__') else str(func)
        try:
            if isinstance(func, Awaitable):
                async def await_func():
                    await func
                background_tasks.create(await_func(), name=f'func {func_name}')
            else:
                result = func()
                if helpers.is_coroutine_function(func) and not isinstance(result, asyncio.Task):
                    async def await_result():
                        await result
                    background_tasks.create(await_result(), name=f'result {func_name}')
        except Exception as e:
            self.handle_exception(e)

    def on_connect(self, handler: Callable | Awaitable) -> None:
        """Called every time a new client connects to NiceGUI.

        The callback has an optional parameter of `nicegui.Client`.
        """
        self._connect_handlers.append(handler)

    def on_disconnect(self, handler: Callable | Awaitable) -> None:
        """Called every time a new client disconnects from NiceGUI.

        The callback has an optional parameter of `nicegui.Client`.

        *Updated in version 3.0.0: The handler is also called when a client reconnects.*
        """
        self._disconnect_handlers.append(handler)

    def on_delete(self, handler: Callable | Awaitable) -> None:
        """Called when a client is deleted.

        The callback has an optional parameter of `nicegui.Client`.

        *Added in version 3.0.0*
        """
        self._delete_handlers.append(handler)

    def on_startup(self, handler: Callable | Awaitable) -> None:
        """Called when NiceGUI is started or restarted.

        Needs to be called before `ui.run()`.
        """
        if self.is_started:
            if core.script_mode:
                raise RuntimeError('Unable to register a startup in script mode. Use a `@ui.page` function instead.')
            raise RuntimeError('Unable to register another startup handler. NiceGUI has already been started.')
        self._startup_handlers.append(handler)

    def on_shutdown(self, handler: Callable | Awaitable) -> None:
        """Called when NiceGUI is shut down or restarted.

        When NiceGUI is shut down or restarted, all tasks still in execution will be automatically canceled.
        """
        self._shutdown_handlers.append(handler)

    def on_exception(self, handler: Callable) -> None:
        """Called when an exception occurs.

        The callback has an optional parameter of `Exception`.
        """
        self._exception_handlers.append(handler)

    def handle_exception(self, exception: Exception) -> None:
        """Handle an exception by invoking all registered exception handlers."""
        if context.slot_stack and context.client is not None:
            context.client.handle_exception(exception)
        for handler in self._exception_handlers:
            result = handler() if not inspect.signature(handler).parameters else handler(exception)
            if helpers.is_coroutine_function(handler):
                background_tasks.create(result, name=f'exception {handler.__name__}')

    def on_page_exception(self, handler: Callable) -> None:
        """Called when an exception occurs in a page and allows to create a custom error page.

        The callback can accept an optional ``Exception`` as argument.
        All UI elements created in the callback are displayed on the error page.
        Asynchronous handlers are currently not supported.

        *Added in version 2.20.0*
        """
        self._page_exception_handler = handler

    def shutdown(self) -> None:
        """Shut down NiceGUI.

        This will programmatically stop the server.
        """
        if self.native.main_window:
            self.native.main_window.destroy()
        if self.config.reload or Server.instance.config.should_reload:
            os.kill(os.getppid(), getattr(signal, 'CTRL_C_EVENT' if platform.system() == 'Windows' else 'SIGINT'))
        else:
            Server.instance.should_exit = True

    def add_static_files(self,
                         url_path: str,
                         local_directory: str | Path,
                         *,
                         follow_symlink: bool = False,
                         max_cache_age: int = 3600) -> None:
        """Add a directory of static files.

        `add_static_files()` makes a local directory available at the specified endpoint, e.g. `'/static'`.
        This is useful for providing local data like images to the frontend.
        Otherwise the browser would not be able to access the files.
        Do only put non-security-critical files in there, as they are accessible to everyone.

        To make a single file accessible, you can use `add_static_file()`.
        For media files which should be streamed, you can use `add_media_files()` or `add_media_file()` instead.

        :param url_path: string that starts with a slash "/" and identifies the path at which the files should be served
        :param local_directory: local folder with files to serve as static content
        :param follow_symlink: whether to follow symlinks (default: False)
        :param max_cache_age: value for max-age set in Cache-Control header (*added in version 2.8.0*)
        """
        if url_path == '/':
            raise ValueError('''Path cannot be "/", because it would hide NiceGUI's internal "/_nicegui" route.''')
        if max_cache_age < 0:
            raise ValueError('''Value of max_cache_age must be a positive integer or 0.''')

        handler = CacheControlledStaticFiles(
            directory=local_directory, follow_symlink=follow_symlink, max_cache_age=max_cache_age)

        @self.get(url_path.rstrip('/') + '/{path:path}')  # NOTE: prevent double slashes in route pattern
        async def static_file(request: Request, path: str = '') -> Response:
            return await handler.get_response(path, request.scope)

    def add_static_file(self, *,
                        local_file: str | Path,
                        url_path: str | None = None,
                        single_use: bool = False,
                        strict: bool = True,
                        max_cache_age: int = 3600) -> str:
        """Add a single static file.

        Allows a local file to be accessed online with enabled caching.
        If `url_path` is not specified, a path will be generated.

        To make a whole folder of files accessible, use `add_static_files()` instead.
        For media files which should be streamed, you can use `add_media_files()` or `add_media_file()` instead.

        :param local_file: local file to serve as static content
        :param url_path: string that starts with a slash "/" and identifies the path at which the file should be served (default: None -> auto-generated URL path)
        :param single_use: whether to remove the route after the file has been downloaded once (default: False)
        :param strict: whether to raise a ``FileNotFoundError`` if the file does not exist (default: True, *added in version 2.12.0*)
        :param max_cache_age: value for max-age set in Cache-Control header (*added in version 2.8.0*)
        :return: encoded URL which can be used to access the file
        """
        if max_cache_age < 0:
            raise ValueError('''Value of max_cache_age must be a positive integer or 0.''')

        file = Path(local_file).resolve()
        if strict and not file.is_file():
            raise FileNotFoundError(f'File not found: {file}')
        path = f'/_nicegui/auto/static/{helpers.hash_file_path(file)}/{file.name}' if url_path is None else url_path

        @self.get(path)
        def read_item() -> FileResponse:
            if single_use:
                self.remove_route(path)
            return FileResponse(file, headers={'Cache-Control': f'public, max-age={max_cache_age}'})

        return urllib.parse.quote(path)

    def add_media_files(self, url_path: str, local_directory: str | Path) -> None:
        """Add directory of media files.

        `add_media_files()` allows a local files to be streamed from a specified endpoint, e.g. `'/media'`.
        This should be used for media files to support proper streaming.
        Otherwise the browser would not be able to access and load the the files incrementally or jump to different positions in the stream.
        Do only put non-security-critical files in there, as they are accessible to everyone.

        To make a single file accessible via streaming, you can use `add_media_file()`.
        For small static files, you can use `add_static_files()` or `add_static_file()` instead.

        :param url_path: string that starts with a slash "/" and identifies the path at which the files should be served
        :param local_directory: local folder with files to serve as media content
        """
        @self.get(url_path.rstrip('/') + '/{filename:path}')  # NOTE: prevent double slashes in route pattern
        def read_item(request: Request, filename: str, nicegui_chunk_size: int = 8192) -> Response:
            local_dir = Path(local_directory).resolve()
            filepath = (local_dir / filename).resolve()
            if not filepath.is_relative_to(local_dir) or not filepath.is_file():
                raise HTTPException(status_code=404, detail='Not Found')
            return get_range_response(filepath, request, chunk_size=nicegui_chunk_size)

    def add_media_file(self, *,
                       local_file: str | Path,
                       url_path: str | None = None,
                       single_use: bool = False,
                       strict: bool = True) -> str:
        """Add a single media file.

        Allows a local file to be streamed.
        If `url_path` is not specified, a path will be generated.

        To make a whole folder of media files accessible via streaming, use `add_media_files()` instead.
        For small static files, you can use `add_static_files()` or `add_static_file()` instead.

        :param local_file: local file to serve as media content
        :param url_path: string that starts with a slash "/" and identifies the path at which the file should be served (default: None -> auto-generated URL path)
        :param single_use: whether to remove the route after the media file has been downloaded once (default: False)
        :param strict: whether to raise a ``FileNotFoundError`` if the file does not exist (default: True, *added in version 2.12.0*)
        :return: encoded URL which can be used to access the file
        """
        file = Path(local_file).resolve()
        if strict and not file.is_file():
            raise FileNotFoundError(f'File not found: {file}')
        path = f'/_nicegui/auto/media/{helpers.hash_file_path(file)}/{file.name}' if url_path is None else url_path

        @self.get(path)
        def read_item(request: Request, nicegui_chunk_size: int = 8192) -> Response:
            if single_use:
                self.remove_route(path)
            return get_range_response(file, request, chunk_size=nicegui_chunk_size)

        return urllib.parse.quote(path)

    def colors(self, *,
               primary: str = '#5898d4',
               secondary: str = '#26a69a',
               accent: str = '#9c27b0',
               dark: str = '#1d1d1d',
               dark_page: str = '#121212',
               positive: str = '#21ba45',
               negative: str = '#c10015',
               info: str = '#31ccec',
               warning: str = '#f2c037',
               **custom_colors: str) -> None:
        """Color Theming

        Sets the main colors (primary, secondary, accent, ...) used by `Quasar <https://quasar.dev/style/theme-builder>`_ on an application-wide basis.

        Note: Use ``ui.colors()`` if you want to set colors after a page has been rendered on a per-page basis.

        *Added in version 3.6.0*

        :param primary: Primary color (default: "#5898d4")
        :param secondary: Secondary color (default: "#26a69a")
        :param accent: Accent color (default: "#9c27b0")
        :param dark: Dark color (default: "#1d1d1d")
        :param dark_page: Dark page color (default: "#121212")
        :param positive: Positive color (default: "#21ba45")
        :param negative: Negative color (default: "#c10015")
        :param info: Info color (default: "#31ccec")
        :param warning: Warning color (default: "#f2c037")
        :param custom_colors: Custom color definitions for branding
        """
        brand: dict[str, str] = self.config.quasar_config['brand']
        brand['primary'] = primary
        brand['secondary'] = secondary
        brand['accent'] = accent
        brand['dark'] = dark
        brand['dark-page'] = dark_page
        brand['positive'] = positive
        brand['negative'] = negative
        brand['info'] = info
        brand['warning'] = warning
        brand.update({name.replace('_', '-'): value for name, value in custom_colors.items()})
        QUASAR_COLORS.update({name.replace('_', '-') for name in custom_colors})

    def remove_route(self, path: str) -> None:
        """Remove routes with the given path."""
        self.routes[:] = [r for r in self.routes if getattr(r, 'path', None) != path]

    def reset(self) -> None:
        """Reset app to its initial state. (Useful for testing.)"""
        self.storage.clear()
        self._startup_handlers.clear()
        self._shutdown_handlers.clear()
        self._connect_handlers.clear()
        self._disconnect_handlers.clear()
        self._delete_handlers.clear()
        self._exception_handlers[:] = [log.exception]
        self._page_exception_handler = None
        self.config = AppConfig()
        self.colors()  # reset colors to default

    @staticmethod
    def clients(path: str | None = None) -> Iterator[Client]:
        """Iterate over clients, with a matching path if provided.

        When using ``@ui.page("/path")``, each client gets a private view of this page.
        Updates must be sent to each client individually, which this iterator simplifies.
        If ``path`` is ``None`` (default), it yields all clients regardless of their path.

        *Added in version 2.7.0*

        :param path: string to filter clients by (or ``None`` to get all clients, *added in version 3.9.0*)
        """
        for client in Client.instances.values():
            if path is None or client.page.path == path:
                yield client



from typing import cast

from fastapi import Request
from starlette.datastructures import UploadFile
from typing_extensions import Self

from ..defaults import DEFAULT_PROP, resolve_defaults
from ..events import Handler, MultiUploadEventArguments, UiEventArguments, UploadEventArguments, handle_event
from ..nicegui import app
from .mixins.disableable_element import DisableableElement
from .mixins.label_element import LabelElement
from .upload_files import create_file_upload


class Upload(LabelElement, DisableableElement, component='upload.js'):
    # pylint: disable=import-outside-toplevel
    from .upload_files import FileUpload, LargeFileUpload, SmallFileUpload

    @resolve_defaults
    def __init__(self, *,
                 multiple: bool = DEFAULT_PROP | False,
                 max_file_size: int | None = DEFAULT_PROP | None,
                 max_total_size: int | None = DEFAULT_PROP | None,
                 max_files: int | None = DEFAULT_PROP | None,
                 on_begin_upload: Handler[UiEventArguments] | None = None,
                 on_upload: Handler[UploadEventArguments] | None = None,
                 on_multi_upload: Handler[MultiUploadEventArguments] | None = None,
                 on_rejected: Handler[UiEventArguments] | None = None,
                 label: str = DEFAULT_PROP | '',
                 auto_upload: bool = DEFAULT_PROP | False,
                 ) -> None:
        """File Upload

        Based on Quasar's `QUploader <https://quasar.dev/vue-components/uploader>`_ component.

        Upload event handlers are called in the following order:

        1. ``on_begin_upload``: The client begins uploading one or more files to the server.
        2. ``on_upload``: The upload of an individual file is complete.
        3. ``on_multi_upload``: The upload of all selected files is complete.

        The following event handler is already called during the file selection process:

        - ``on_rejected``: One or more files have been rejected.

        :param multiple: allow uploading multiple files at once (default: `False`)
        :param max_file_size: maximum file size in bytes (default: `0`)
        :param max_total_size: maximum total size of all files in bytes (default: `0`)
        :param max_files: maximum number of files (default: `0`)
        :param on_begin_upload: callback to execute when upload begins  (*added in version 2.14.0*)
        :param on_upload: callback to execute for each uploaded file
        :param on_multi_upload: callback to execute after multiple files have been uploaded
        :param on_rejected: callback to execute when one or more files have been rejected during file selection
        :param label: label for the uploader (default: `''`)
        :param auto_upload: automatically upload files when they are selected (default: `False`)
        """
        super().__init__(label=label)
        self._props.set_bool('multiple', multiple)
        self._props.set_bool('auto-upload', auto_upload)
        self._props['url'] = f'/_nicegui/client/{self.client.id}/upload/{self.id}'

        self._props.set_optional('max-file-size', max_file_size)
        self._props.set_optional('max-total-size', max_total_size)
        self._props.set_optional('max-files', max_files)

        if multiple and on_multi_upload:
            self._props['batch'] = True

        self._begin_upload_handlers = [on_begin_upload] if on_begin_upload else []
        self._upload_handlers = [on_upload] if on_upload else []
        self._multi_upload_handlers = [on_multi_upload] if on_multi_upload else []

        @app.post(self._props['url'])
        async def upload_route(request: Request) -> dict[str, str]:
            for begin_upload_handler in self._begin_upload_handlers:
                handle_event(begin_upload_handler, UiEventArguments(sender=self, client=self.client))
            async with request.form() as form:
                files = [await create_file_upload(cast(UploadFile, data)) for data in form.values()]
            await self.handle_uploads(files)
            return {'upload': 'success'}

        if on_rejected:
            self.on_rejected(on_rejected)

    async def handle_uploads(self, files: list[FileUpload]) -> None:
        """Handle the uploaded files.

        This method is primarily intended for internal use and for simulating file uploads in tests.
        """
        assert all(isinstance(f, Upload.FileUpload) for f in files), \
            'since NiceGUI 3.0, uploads must be a list of FileUpload instances'
        for file in files:
            for upload_handler in self._upload_handlers:
                handle_event(upload_handler, UploadEventArguments(sender=self, client=self.client, file=file))
        multi_upload_args = MultiUploadEventArguments(sender=self, client=self.client, files=files)
        for multi_upload_handler in self._multi_upload_handlers:
            handle_event(multi_upload_handler, multi_upload_args)

    def on_begin_upload(self, callback: Handler[UiEventArguments]) -> Self:
        """Add a callback to be invoked when the upload begins."""
        self._begin_upload_handlers.append(callback)
        return self

    def on_upload(self, callback: Handler[UploadEventArguments]) -> Self:
        """Add a callback to be invoked when a file is uploaded."""
        self._upload_handlers.append(callback)
        return self

    def on_multi_upload(self, callback: Handler[MultiUploadEventArguments]) -> Self:
        """Add a callback to be invoked when multiple files have been uploaded."""
        self._multi_upload_handlers.append(callback)
        return self

    def on_rejected(self, callback: Handler[UiEventArguments]) -> Self:
        """Add a callback to be invoked when one or more files have been rejected during file selection."""
        self.on('rejected', lambda: handle_event(callback, UiEventArguments(sender=self, client=self.client)), args=[])
        return self

    def reset(self) -> None:
        """Clear the upload queue."""
        self.run_method('reset')

    def _handle_delete(self) -> None:
        app.remove_route(self._props['url'])
        super()._handle_delete()



from pathlib import Path

import aiofiles

from .. import background_tasks, core, json
from ..logging import log
from .persistent_dict import PersistentDict


class FilePersistentDict(PersistentDict):

    def __init__(self, filepath: Path, encoding: str | None = None, *, indent: bool = False) -> None:
        self.filepath = filepath
        self.encoding = encoding
        self.indent = indent
        super().__init__(data={}, on_change=self.backup)

    async def initialize(self) -> None:
        try:
            if self.filepath.exists():
                async with aiofiles.open(self.filepath, encoding=self.encoding) as f:
                    data = json.loads(await f.read())
            else:
                data = {}
            self.update(data)
        except Exception:
            log.warning(f'Could not load storage file {self.filepath}')

    def initialize_sync(self) -> None:
        try:
            if self.filepath.exists():
                data = json.loads(self.filepath.read_text(encoding=self.encoding))
            else:
                data = {}
            self.update(data)
        except Exception:
            log.warning(f'Could not load storage file {self.filepath}')

    def backup(self) -> None:
        """Back up the data to the given file path."""
        if not self.filepath.exists():
            if not self:
                return
            self.filepath.parent.mkdir(exist_ok=True)

        @background_tasks.await_on_shutdown
        async def async_backup() -> None:
            async with aiofiles.open(self.filepath, 'w', encoding=self.encoding) as f:
                await f.write(json.dumps(self, indent=self.indent))

        if core.loop and core.loop.is_running():
            background_tasks.create_lazy(async_backup(), name=self.filepath.stem)
        else:
            self.filepath.write_text(json.dumps(self, indent=self.indent), encoding=self.encoding)

    def clear(self) -> None:
        super().clear()
        self.filepath.unlink(missing_ok=True)



from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.value_element import ValueElement


class Slider(ValueElement, DisableableElement):

    @resolve_defaults
    def __init__(self, *,
                 min: float,  # pylint: disable=redefined-builtin
                 max: float,  # pylint: disable=redefined-builtin
                 step: float = DEFAULT_PROP | 1.0,
                 value: float | None = DEFAULT_PROPS['model-value'] | None,
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        """Slider

        This element is based on Quasar's `QSlider <https://quasar.dev/vue-components/slider>`_ component.

        :param min: lower bound of the slider
        :param max: upper bound of the slider
        :param step: step size
        :param value: initial value to set position of the slider
        :param on_change: callback which is invoked when the user releases the slider
        """
        super().__init__(tag='q-slider', value=value, on_value_change=on_change, throttle=0.05)
        self._props['min'] = min
        self._props['max'] = max
        self._props['step'] = step



from collections.abc import Callable
from typing import Any, cast

from typing_extensions import Self

from ...binding import BindableProperty, bind, bind_from, bind_to
from ...element import Element


class DisableableElement(Element):
    enabled = BindableProperty(
        on_change=lambda sender, value: cast(Self, sender)._handle_enabled_change(value))  # pylint: disable=protected-access

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.enabled = True
        self.ignores_events_when_disabled = True

    @property
    def is_ignoring_events(self) -> bool:
        """Return whether the element is currently ignoring events."""
        if super().is_ignoring_events:
            return True
        return not self.enabled and self.ignores_events_when_disabled

    def enable(self) -> None:
        """Enable the element."""
        self.enabled = True

    def disable(self) -> None:
        """Disable the element."""
        self.enabled = False

    def bind_enabled_to(self,
                        target_object: Any,
                        target_name: str = 'enabled',
                        forward: Callable[[Any], Any] | None = None, *,
                        strict: bool | None = None,
                        ) -> Self:
        """Bind the enabled state of this element to the target object's target_name property.

        The binding works one way only, from this element to the target.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_to(self, 'enabled', target_object, target_name, forward, self_strict=False, other_strict=strict)
        return self

    def bind_enabled_from(self,
                          target_object: Any,
                          target_name: str = 'enabled',
                          backward: Callable[[Any], Any] | None = None, *,
                          strict: bool | None = None,
                          ) -> Self:
        """Bind the enabled state of this element from the target object's target_name property.

        The binding works one way only, from the target to this element.
        The update happens immediately and whenever a value changes.

        :param target_object: The object to bind from.
        :param target_name: The name of the property to bind from.
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind_from(self, 'enabled', target_object, target_name, backward, self_strict=False, other_strict=strict)
        return self

    def bind_enabled(self,
                     target_object: Any,
                     target_name: str = 'enabled', *,
                     forward: Callable[[Any], Any] | None = None,
                     backward: Callable[[Any], Any] | None = None,
                     strict: bool | None = None,
                     ) -> Self:
        """Bind the enabled state of this element to the target object's target_name property.

        The binding works both ways, from this element to the target and from the target to this element.
        The update happens immediately and whenever a value changes.
        The backward binding takes precedence for the initial synchronization.

        :param target_object: The object to bind to.
        :param target_name: The name of the property to bind to.
        :param forward: A function to apply to the value before applying it to the target (default: identity).
        :param backward: A function to apply to the value before applying it to this element (default: identity).
        :param strict: Whether to check (and raise) if the target object has the specified property (default: None,
            performs a check if the object is not a dictionary, *added in version 3.0.0*).
        """
        bind(self, 'enabled', target_object, target_name,
             forward=forward, backward=backward,
             self_strict=False, other_strict=strict)
        return self

    def set_enabled(self, value: bool) -> None:
        """Set the enabled state of the element."""
        self.enabled = value

    def _handle_enabled_change(self, enabled: bool) -> None:
        """Called when the element is enabled or disabled.

        :param enabled: The new state.
        """
        self._props['disable'] = not enabled



__all__ = [
    'add_body_html',
    'add_css',
    'add_head_html',
    'add_sass',
    'add_scss',
    'aggrid',
    'altair',
    'anywidget',
    'audio',
    'avatar',
    'badge',
    'button',
    'button_group',
    'card',
    'card_actions',
    'card_section',
    'carousel',
    'carousel_slide',
    'chat_message',
    'checkbox',
    'chip',
    'circular_progress',
    'clipboard',
    'code',
    'codemirror',
    'color_input',
    'color_picker',
    'colors',
    'column',
    'context',
    'context_menu',
    'dark_mode',
    'date',
    'date_input',
    'dialog',
    'download',
    'drawer',
    'dropdown_button',
    'echart',
    'editor',
    'element',
    'expansion',
    'fab',
    'fab_action',
    'footer',
    'fullscreen',
    'grid',
    'header',
    'highchart',
    'html',
    'icon',
    'image',
    'input',
    'input_chips',
    'interactive_image',
    'item',
    'item_label',
    'item_section',
    'joystick',
    'json_editor',
    'keyboard',
    'knob',
    'label',
    'leaflet',
    'left_drawer',
    'line_plot',
    'linear_progress',
    'link',
    'link_target',
    'list',
    'log',
    'markdown',
    'matplotlib',
    'menu',
    'menu_item',
    'mermaid',
    'navigate',
    'notification',
    'notify',
    'number',
    'on',
    'on_exception',
    'page',
    'page_scroller',
    'page_sticky',
    'page_title',
    'pagination',
    'parallax',
    'plotly',
    'pyplot',
    'query',
    'radio',
    'range',
    'rating',
    'refreshable',
    'refreshable_method',
    'restructured_text',
    'right_drawer',
    'row',
    'run',
    'run_javascript',
    'run_with',
    'scene',
    'scene_view',
    'scroll_area',
    'select',
    'separator',
    'skeleton',
    'slide_item',
    'slider',
    'space',
    'spinner',
    'splitter',
    'state',
    'step',
    'stepper',
    'stepper_navigation',
    'sub_pages',
    'switch',
    'tab',
    'tab_panel',
    'tab_panels',
    'table',
    'tabs',
    'teleport',
    'textarea',
    'time',
    'time_input',
    'timeline',
    'timeline_entry',
    'timer',
    'toggle',
    'tooltip',
    'tree',
    'update',
    'upload',
    'video',
    'xterm',
]

from .context import context
from .element import Element as element
from .elements.aggrid import AgGrid as aggrid
from .elements.altair import Altair as altair
from .elements.anywidget import AnyWidget as anywidget
from .elements.audio import Audio as audio
from .elements.avatar import Avatar as avatar
from .elements.badge import Badge as badge
from .elements.button import Button as button
from .elements.button_dropdown import DropdownButton as dropdown_button
from .elements.button_group import ButtonGroup as button_group
from .elements.card import Card as card
from .elements.card import CardActions as card_actions
from .elements.card import CardSection as card_section
from .elements.carousel import Carousel as carousel
from .elements.carousel import CarouselSlide as carousel_slide
from .elements.chat_message import ChatMessage as chat_message
from .elements.checkbox import Checkbox as checkbox
from .elements.chip import Chip as chip
from .elements.code import Code as code
from .elements.codemirror import CodeMirror as codemirror
from .elements.color_input import ColorInput as color_input
from .elements.color_picker import ColorPicker as color_picker
from .elements.colors import Colors as colors
from .elements.column import Column as column
from .elements.context_menu import ContextMenu as context_menu
from .elements.dark_mode import DarkMode as dark_mode
from .elements.date import Date as date
from .elements.date_input import DateInput as date_input
from .elements.dialog import Dialog as dialog
from .elements.drawer import Drawer as drawer
from .elements.drawer import LeftDrawer as left_drawer
from .elements.drawer import RightDrawer as right_drawer
from .elements.echart import EChart as echart
from .elements.editor import Editor as editor
from .elements.expansion import Expansion as expansion
from .elements.fab import Fab as fab
from .elements.fab import FabAction as fab_action
from .elements.footer import Footer as footer
from .elements.fullscreen import Fullscreen as fullscreen
from .elements.grid import Grid as grid
from .elements.header import Header as header
from .elements.highchart import highchart
from .elements.html import Html as html
from .elements.icon import Icon as icon
from .elements.image import Image as image
from .elements.input import Input as input  # pylint: disable=redefined-builtin
from .elements.input_chips import InputChips as input_chips
from .elements.interactive_image import InteractiveImage as interactive_image
from .elements.item import Item as item
from .elements.item import ItemLabel as item_label
from .elements.item import ItemSection as item_section
from .elements.joystick import Joystick as joystick
from .elements.json_editor import JsonEditor as json_editor
from .elements.keyboard import Keyboard as keyboard
from .elements.knob import Knob as knob
from .elements.label import Label as label
from .elements.leaflet import Leaflet as leaflet
from .elements.line_plot import LinePlot as line_plot
from .elements.link import Link as link
from .elements.link import LinkTarget as link_target
from .elements.list import List as list  # pylint: disable=redefined-builtin
from .elements.log import Log as log
from .elements.markdown import Markdown as markdown
from .elements.menu import Menu as menu
from .elements.menu import MenuItem as menu_item
from .elements.mermaid import Mermaid as mermaid
from .elements.notification import Notification as notification
from .elements.number import Number as number
from .elements.page_scroller import PageScroller as page_scroller
from .elements.page_sticky import PageSticky as page_sticky
from .elements.pagination import Pagination as pagination
from .elements.parallax import Parallax as parallax
from .elements.plotly import Plotly as plotly
from .elements.progress import CircularProgress as circular_progress
from .elements.progress import LinearProgress as linear_progress
from .elements.pyplot import Matplotlib as matplotlib
from .elements.pyplot import Pyplot as pyplot
from .elements.query import Query as query
from .elements.radio import Radio as radio
from .elements.range import Range as range  # pylint: disable=redefined-builtin
from .elements.rating import Rating as rating
from .elements.restructured_text import ReStructuredText as restructured_text
from .elements.row import Row as row
from .elements.scene import Scene as scene
from .elements.scene import SceneView as scene_view
from .elements.scroll_area import ScrollArea as scroll_area
from .elements.select import Select as select
from .elements.separator import Separator as separator
from .elements.skeleton import Skeleton as skeleton
from .elements.slide_item import SlideItem as slide_item
from .elements.slider import Slider as slider
from .elements.space import Space as space
from .elements.spinner import Spinner as spinner
from .elements.splitter import Splitter as splitter
from .elements.stepper import Step as step
from .elements.stepper import Stepper as stepper
from .elements.stepper import StepperNavigation as stepper_navigation
from .elements.sub_pages import SubPages as sub_pages
from .elements.switch import Switch as switch
from .elements.table import Table as table
from .elements.tabs import Tab as tab
from .elements.tabs import TabPanel as tab_panel
from .elements.tabs import TabPanels as tab_panels
from .elements.tabs import Tabs as tabs
from .elements.teleport import Teleport as teleport
from .elements.textarea import Textarea as textarea
from .elements.time import Time as time
from .elements.time_input import TimeInput as time_input
from .elements.timeline import Timeline as timeline
from .elements.timeline import TimelineEntry as timeline_entry
from .elements.timer import Timer as timer
from .elements.toggle import Toggle as toggle
from .elements.tooltip import Tooltip as tooltip
from .elements.tree import Tree as tree
from .elements.upload import Upload as upload
from .elements.video import Video as video
from .elements.xterm import Xterm as xterm
from .functions import clipboard
from .functions.download import download
from .functions.html import add_body_html, add_head_html
from .functions.javascript import run_javascript
from .functions.navigate import navigate
from .functions.notify import notify
from .functions.on import on
from .functions.on_exception import on_exception
from .functions.page_title import page_title
from .functions.refreshable import refreshable, refreshable_method, state
from .functions.style import add_css, add_sass, add_scss
from .functions.update import update
from .page import page
from .ui_run import run
from .ui_run_with import run_with



from collections.abc import Callable

from .mixins.content_element import ContentElement


class Html(ContentElement, component='html.js'):

    def __init__(self, content: str = '', *, sanitize: Callable[[str], str] | bool = True, tag: str = 'div') -> None:
        """HTML Element

        Renders arbitrary HTML onto the page, wrapped in the specified tag.
        `Tailwind <https://tailwindcss.com/>`_ can be used for styling.
        You can also use `ui.add_head_html` to add html code into the head of the document and `ui.add_body_html`
        to add it into the body.

        :param content: the HTML code to be displayed
        :param sanitize: sanitization mode:
            ``True`` (default) uses client-side sanitization via DOMPurify,
            ``False`` disables sanitization (use only with trusted content),
            or pass a callable to apply server-side sanitization
        :param tag: the HTML tag to wrap the content in (default: "div")
        """
        self._sanitize = sanitize
        super().__init__(content=content)
        self._props['tag'] = tag
        self._props['sanitize'] = sanitize is True

    def _handle_content_change(self, content: str) -> None:
        if callable(self._sanitize):
            content = self._sanitize(content)
        super()._handle_content_change(content)



from ..defaults import DEFAULT_PROP, resolve_defaults
from .label import Label as label
from .mixins.color_elements import TextColorElement
from .mixins.value_element import ValueElement


class LinearProgress(ValueElement, TextColorElement):
    VALUE_PROP = 'value'

    @resolve_defaults
    def __init__(self,
                 value: float = DEFAULT_PROP | 0.0, *,
                 size: str | None = DEFAULT_PROP | None,
                 show_value: bool = True,
                 color: str | None = DEFAULT_PROP | 'primary',
                 ) -> None:
        """Linear Progress

        A linear progress bar wrapping Quasar's
        `QLinearProgress <https://quasar.dev/vue-components/linear-progress>`_ component.

        :param value: the initial value of the field (from 0.0 to 1.0)
        :param size: the height of the progress bar (default: "20px" with value label and "4px" without)
        :param show_value: whether to show a value label in the center (default: `True`)
        :param color: color (either a Quasar, Tailwind, or CSS color or `None`, default: "primary")
        """
        super().__init__(tag='q-linear-progress', value=value, on_value_change=None, text_color=color)
        self._props['size'] = size if size is not None else '20px' if show_value else '4px'

        if show_value:
            with self:
                label().classes('absolute-center text-sm text-white').bind_text_from(self, 'value')


class CircularProgress(ValueElement, TextColorElement):
    VALUE_PROP = 'value'

    @resolve_defaults
    def __init__(self,
                 value: float = DEFAULT_PROP | 0.0, *,
                 min: float = DEFAULT_PROP | 0.0,  # pylint: disable=redefined-builtin
                 max: float = DEFAULT_PROP | 1.0,  # pylint: disable=redefined-builtin
                 size: str = DEFAULT_PROP | 'xl',
                 show_value: bool = True,
                 color: str | None = DEFAULT_PROP | 'primary',
                 ) -> None:
        """Circular Progress

        A circular progress bar wrapping Quasar's
        `QCircularProgress <https://quasar.dev/vue-components/circular-progress>`_.

        :param value: the initial value of the field
        :param min: the minimum value (default: 0.0)
        :param max: the maximum value (default: 1.0)
        :param size: the size of the progress circle (default: "xl")
        :param show_value: whether to show a value label in the center (default: `True`)
        :param color: color (either a Quasar, Tailwind, or CSS color or `None`, default: "primary")
        """
        super().__init__(tag='q-circular-progress', value=value, on_value_change=None, text_color=color)
        self._props['min'] = min
        self._props['max'] = max
        self._props['size'] = size
        self._props['show-value'] = True  # NOTE always activate the default slot because this is expected by ui.element
        self._props['track-color'] = 'grey-4'

        if show_value:
            with self:
                label().classes('absolute-center text-xs').bind_text_from(self, 'value')



from typing import Any, Literal

from ..context import context

ARG_MAP = {
    'close_button': 'closeBtn',
    'multi_line': 'multiLine',
}


# pylint: disable=unused-argument
def notify(message: Any, *,
           position: Literal[
               'top-left',
               'top-right',
               'bottom-left',
               'bottom-right',
               'top',
               'bottom',
               'left',
               'right',
               'center',
           ] = 'bottom',
           close_button: bool | str = False,
           type: Literal[  # pylint: disable=redefined-builtin
               'positive',
               'negative',
               'warning',
               'info',
               'ongoing',
           ] | None = None,
           color: str | None = None,
           multi_line: bool = False,
           **kwargs: Any,
           ) -> None:
    """Notification

    Displays a notification on the screen.

    :param message: content of the notification
    :param position: position on the screen ("top-left", "top-right", "bottom-left", "bottom-right", "top", "bottom", "left", "right" or "center", default: "bottom")
    :param close_button: optional label of a button to dismiss the notification (default: `False`)
    :param type: optional type ("positive", "negative", "warning", "info" or "ongoing")
    :param color: optional color name
    :param multi_line: enable multi-line notifications

    Note: You can pass additional keyword arguments according to `Quasar's Notify API <https://quasar.dev/quasar-plugins/notify#notify-api>`_.
    """
    options = {ARG_MAP.get(key, key): value for key, value in locals().items() if key != 'kwargs' and value is not None}
    options['message'] = str(message)
    options.update(kwargs)
    client = context.client
    client.outbox.enqueue_message('notify', options, client.id)



from ..client import Client
from ..context import context


def add_head_html(code: str, *, shared: bool = False) -> None:
    """Add HTML code to the head of the page.

    :param code: HTML code to add
    :param shared: if True, the code is added to all pages
    """
    if shared:
        Client.shared_head_html += code + '\n'
    else:
        client = context.client
        if client.has_socket_connection:
            client.run_javascript(f'document.head.insertAdjacentHTML("beforeend", {code!r});')
        client._head_html += code + '\n'  # pylint: disable=protected-access


def add_body_html(code: str, *, shared: bool = False) -> None:
    """Add HTML code to the body of the page.

    :param code: HTML code to add
    :param shared: if True, the code is added to all pages
    """
    if shared:
        Client.shared_body_html += code + '\n'
    else:
        client = context.client
        if client.has_socket_connection:
            client.run_javascript(f'document.querySelector("#app").insertAdjacentHTML("beforebegin", {code!r});')
        client._body_html += code + '\n'  # pylint: disable=protected-access



import functools
import inspect
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, TypeVar

from typing_extensions import ParamSpec

from .element import Element

T = TypeVar('T')

P = ParamSpec('P')
R = TypeVar('R')


@dataclass(kw_only=True, slots=True)
class Sentinel:
    key: str | None

    def __or__(self, other: T) -> T:
        return SentinelValue(key=self.key, default=other)  # type: ignore[return-value]


@dataclass(kw_only=True, slots=True)
class SentinelValue:
    key: str | None
    default: Any


class SentinelFactory:

    def __getitem__(self, prop_key: str) -> Sentinel:
        return Sentinel(key=prop_key)


DEFAULT_PROPS = SentinelFactory()
DEFAULT_PROP = Sentinel(key=None)


def resolve_defaults(original_func: Callable[P, R]) -> Callable[P, R]:
    """This decorator makes the function resolve default properties set via ``default_props``.

    If a parameter has a default value which looks like ``DEFAULT_PROPS['prop-key'] | default_value``,
    the actual value will be taken from the element's ``_default_props`` dictionary with key "prop-key" if present,
    otherwise the specified ``default_value`` is used.

    If a parameter has a default value which looks like ``DEFAULT_PROP | default_value``,
    the dictionary key will be inferred from the parameter name (converting snake_case to kebab-case).
    """
    signature = inspect.signature(original_func)

    @functools.wraps(original_func)
    def decorated(*args: P.args, **kwargs: P.kwargs) -> R:
        bound = signature.bind(*args, **kwargs)
        bound.apply_defaults()

        el: Element = bound.arguments['self']

        for param_name, value in bound.arguments.items():
            if isinstance(value, SentinelValue):
                key = value.key or param_name.replace('_', '-')
                kwargs[param_name] = el._default_props.get(key, value.default)  # pylint: disable=protected-access
        return original_func(*args, **kwargs)

    return decorated



from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .mixins.disableable_element import DisableableElement
from .mixins.value_element import ValueElement


class Rating(ValueElement, DisableableElement):

    @resolve_defaults
    def __init__(self,
                 value: float | None = DEFAULT_PROPS['model-value'] | None,
                 max: int = DEFAULT_PROP | 5,  # pylint: disable=redefined-builtin
                 icon: str | None = DEFAULT_PROP | None,
                 icon_selected: str | None = DEFAULT_PROP | None,
                 icon_half: str | None = DEFAULT_PROP | None,
                 color: str | list[str] | None = DEFAULT_PROP | 'primary',
                 size: str | None = DEFAULT_PROP | None,
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        """Rating

        This element is based on Quasar's `QRating <https://quasar.dev/vue-components/rating>`_ component.

        *Added in version 2.12.0*

        :param value: initial value (default: ``None``)
        :param max: maximum rating, number of icons (default: 5)
        :param icon: name of icons to be displayed (default: star)
        :param icon_selected: name of an icon to be displayed when selected (default: same as ``icon``)
        :param icon_half: name of an icon to be displayed when half-selected (default: same as ``icon``)
        :param color: color(s) of the icons (Quasar, Tailwind, or CSS colors or ``None``, default: "primary")
        :param size: size in CSS units, including unit name or standard size name (xs|sm|md|lg|xl), examples: 16px, 2rem
        :param on_change: callback to execute when selection changes
        """
        super().__init__(tag='q-rating', value=value, on_value_change=on_change)

        self._props['max'] = max
        self._props.set_optional('icon', icon)
        self._props.set_optional('color', color)
        self._props.set_optional('icon-selected', icon_selected)
        self._props.set_optional('icon-half', icon_half)
        self._props.set_optional('size', size)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from .image import Image

if TYPE_CHECKING:
    from PIL.Image import Image as PIL_Image


class Parallax(Image, component='image.js'):

    def __init__(self, source: str | Path | PIL_Image = '', *, height: float = 500.0, speed: float = 1.0) -> None:
        """Parallax Image

        Displays an image with a parallax effect.
        This element is based on Quasar's `Parallax <https://quasar.dev/vue-components/parallax>`_ component.

        *Added in version 3.9.0.*

        :param source: the source of the image; can be a URL, local file path, a base64 string or a PIL image
        :param height: the height of the parallax image in pixels (default: 500.0)
        :param speed: the speed (0 to 1) of the parallax effect (default: 1.0)
        """
        super().__init__(source=source)
        self._props['height'] = height
        self._props['speed'] = speed
        self._props['tag'] = 'q-parallax'



from __future__ import annotations

import asyncio
import re
from collections.abc import Callable
from typing import Any, TypeVar, overload
from uuid import uuid4

import httpx
import socketio

from nicegui import Client, ElementFilter, ui
from nicegui.nicegui import _on_handshake
from nicegui.outbox import Message

from .user_download import UserDownload
from .user_interaction import UserInteraction
from .user_navigate import UserNavigate
from .user_notify import UserNotify

# pylint: disable=protected-access


T = TypeVar('T', bound=ui.element)


class User:
    current_user: User | None = None

    def __init__(self, client: httpx.AsyncClient) -> None:
        self.http_client = client
        self.sio = socketio.AsyncClient()
        self.client: Client | None = None
        self.back_history: list[str] = []
        self.forward_history: list[str] = []
        self.navigate = UserNavigate(self)
        self.notify = UserNotify()
        self.download = UserDownload(self)
        self.tab_id = str(uuid4())
        self.javascript_rules: dict[re.Pattern, Callable[[re.Match], Any]] = {
            re.compile('.*__IS_DRAWER_OPEN__'): lambda _: True,  # see https://github.com/zauberzeug/nicegui/issues/4508
        }

    @property
    def _client(self) -> Client:
        if self.client is None:
            raise ValueError('This user has not opened a page yet. Did you forgot to call .open()?')
        return self.client

    def __enter__(self) -> Client:
        return self._client.__enter__()

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        return self._client.__exit__(exc_type, exc_val, exc_tb)

    def __getattribute__(self, name: str) -> Any:
        if name not in {'notify', 'navigate', 'download'}:  # NOTE: avoid infinite recursion
            ui.navigate = self.navigate
            ui.notify = self.notify
            ui.download = self.download
        return super().__getattribute__(name)

    async def open(self, path: str, *, clear_forward_history: bool = True) -> Client:
        """Open the given path."""
        response = await self.http_client.get(path, follow_redirects=True)
        assert response.status_code == 200, f'Expected status code 200, got {response.status_code}'
        if response.headers.get('X-Nicegui-Content') != 'page':
            raise ValueError(f'Expected a page response, got {response.text}')

        match = re.search(r"'client_id': '([0-9a-f-]+)'", response.text)
        assert match is not None
        client_id = match.group(1)
        self.client = Client.instances[client_id]
        self.sio.on('connect')
        await _on_handshake(f'test-{uuid4()}', {
            'client_id': self.client.id,
            'tab_id': self.tab_id,
            'document_id': str(uuid4()),
        })
        self.back_history.append(path)
        if clear_forward_history:
            self.forward_history.clear()
        self._patch_outbox_emit_function()
        return self.client

    def _patch_outbox_emit_function(self) -> None:
        original_emit = self._client.outbox._emit

        async def simulated_emit(message: Message) -> None:
            await original_emit(message)
            _, type_, data = message
            if type_ == 'run_javascript':
                for rule, result in self.javascript_rules.items():
                    match = rule.match(data['code'])
                    if match:
                        self._client.handle_javascript_response({
                            'request_id': data['request_id'],
                            'result': result(match),
                        })

        self._client.outbox._emit = simulated_emit  # type: ignore

    @overload
    async def should_see(self,
                         target: str | type[T],
                         *,
                         retries: int = 3,
                         ) -> None:
        ...

    @overload
    async def should_see(self,
                         *,
                         kind: type[T] | None = None,
                         marker: str | list[str] | None = None,
                         content: str | list[str] | None = None,
                         retries: int = 3,
                         ) -> None:
        ...

    async def should_see(self,
                         target: str | type[T] | None = None,
                         *,
                         kind: type[T] | None = None,
                         marker: str | list[str] | None = None,
                         content: str | list[str] | None = None,
                         retries: int = 3,
                         ) -> None:
        """Assert that the page contains an element fulfilling certain filter rules.

        Note that there is no scrolling in the user simulation -- the entire page is always *visible*.
        Due to asynchronous execution, sometimes the expected elements only appear after a short delay.

        By default `should_see` makes three attempts to find the element before failing.
        This can be adjusted with the `retries` parameter.
        """
        for _ in range(retries):
            with self._client:
                if self.notify.contains(target) or self._gather_elements(target, kind, marker, content):
                    return
                await asyncio.sleep(0.1)
        raise AssertionError('expected to see at least one ' + self._build_error_message(target, kind, marker, content))

    @overload
    async def should_not_see(self,
                             target: str | type[T],
                             *,
                             retries: int = 3,
                             ) -> None:
        ...

    @overload
    async def should_not_see(self,
                             *,
                             kind: type[T] | None = None,
                             marker: str | list[str] | None = None,
                             content: str | list[str] | None = None,
                             retries: int = 3,
                             ) -> None:
        ...

    async def should_not_see(self,
                             target: str | type[T] | None = None,
                             *,
                             kind: type[T] | None = None,
                             marker: str | list[str] | None = None,
                             content: str | list[str] | None = None,
                             retries: int = 3,
                             ) -> None:
        """Assert that the page does not contain an input with the given value."""
        for _ in range(retries):
            with self._client:
                if not self.notify.contains(target) and not self._gather_elements(target, kind, marker, content):
                    return
                await asyncio.sleep(0.05)
        raise AssertionError('expected not to see any ' + self._build_error_message(target, kind, marker, content))

    @overload
    def find(self,
             target: str,
             ) -> UserInteraction[ui.element]:
        ...

    @overload
    def find(self,
             target: type[T],
             ) -> UserInteraction[T]:
        ...

    @overload
    def find(self: User,
             *,
             marker: str | list[str] | None = None,
             content: str | list[str] | None = None,
             ) -> UserInteraction[ui.element]:
        ...

    @overload
    def find(self,
             *,
             kind: type[T],
             marker: str | list[str] | None = None,
             content: str | list[str] | None = None,
             ) -> UserInteraction[T]:
        ...

    def find(self,
             target: str | type[T] | None = None,
             *,
             kind: type[T] | None = None,
             marker: str | list[str] | None = None,
             content: str | list[str] | None = None,
             ) -> UserInteraction[T]:
        """Select elements for interaction."""
        with self._client:
            elements = self._gather_elements(target, kind, marker, content)
            if not elements:
                raise AssertionError('expected to find at least one ' +
                                     self._build_error_message(target, kind, marker, content))
        return UserInteraction(self, elements, target)

    @property
    def current_layout(self) -> ui.element:
        """Return the root layout element of the current page."""
        return self._client.layout

    def _gather_elements(
        self,
        target: str | type[T] | None = None,
        kind: type[T] | None = None,
        marker: str | list[str] | None = None,
        content: str | list[str] | None = None,
    ) -> set[T]:
        if target is None:
            if kind is None:
                elements = set(ElementFilter(marker=marker, content=content, only_visible=True))
            else:
                elements = set(ElementFilter(kind=kind, marker=marker, content=content, only_visible=True))
        elif isinstance(target, str):
            elements = set(ElementFilter(marker=target, only_visible=True)) \
                .union(ElementFilter(content=target, only_visible=True))
        else:
            elements = set(ElementFilter(kind=target, only_visible=True))
        return elements  # type: ignore

    def _build_error_message(self,
                             target: str | type[T] | None = None,
                             kind: type[T] | None = None,
                             marker: str | list[str] | None = None,
                             content: str | list[str] | None = None,
                             ) -> str:
        if isinstance(target, str):
            return f'element with marker={target} or content={target} on the page:\n{self.current_layout}'
        elif target is not None:
            return f'element of type {target.__name__} on the page:\n{self.current_layout}'
        elif kind is not None:
            return f'element of type {kind.__name__} with {marker=} and {content=} on the page:\n{self.current_layout}'
        else:
            return f'element with {marker=} and {content=} on the page:\n{self.current_layout}'



from typing import Literal

from ..element import Element


class Row(Element, default_classes='nicegui-row'):

    def __init__(self, *,
                 wrap: bool = True,
                 align_items: Literal['start', 'end', 'center', 'baseline', 'stretch'] | None = None,
                 ) -> None:
        """Row Element

        Provides a container which arranges its child in a row.

        :param wrap: whether to wrap the content (default: `True`)
        :param align_items: alignment of the items in the row ("start", "end", "center", "baseline", or "stretch"; default: `None`)
        """
        super().__init__('div')
        self._classes.append('row')  # NOTE: for compatibility with Quasar's col-* classes
        if align_items:
            self._classes.append(f'items-{align_items}')

        if not wrap:
            self._style['flex-wrap'] = 'nowrap'



from __future__ import annotations

import asyncio
import gzip
import json
import logging
import re
from collections.abc import AsyncIterator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any
from uuid import uuid4

import socketio
import socketio.exceptions

from . import background_tasks, core, helpers
from .client import Client
from .logging import log
from .timer import Timer as timer

if TYPE_CHECKING:
    import httpx

RELAY_HOST = 'https://on-air.nicegui.io/'


@dataclass(kw_only=True, slots=True)
class Stream:
    data: AsyncIterator[bytes]
    response: httpx.Response


class Air:

    def __init__(self, token: str) -> None:
        import httpx  # pylint: disable=import-outside-toplevel

        self.log = logging.getLogger('nicegui.air')
        self.token = token
        self.relay = socketio.AsyncClient()
        self.client = httpx.AsyncClient(transport=httpx.ASGITransport(app=core.app))
        self.streaming_client = httpx.AsyncClient()
        self.connecting = False
        self.streams: dict[str, Stream] = {}
        self.remote_url: str | None = None
        self._host_unreachable_warning = f'On Air host "{RELAY_HOST}" is not reachable. Please check your internet connection.'

        timer(5, self.connect)  # ensure we stay connected

        @self.relay.on('http')
        async def _handle_http(data: dict[str, Any]) -> dict[str, Any]:
            headers: dict[str, Any] = data['headers']
            headers.update({'Accept-Encoding': 'identity', 'X-Forwarded-Prefix': data['prefix']})
            url = 'http://test' + data['path']
            request = self.client.build_request(
                data['method'],
                url,
                params=data['params'],
                headers=headers,
                content=data['body'],
            )
            response = await self.client.send(request)
            self.client.cookies.clear()
            instance_id = data['instance-id']
            content = response.content.replace(
                b'const extraHeaders = {};',
                (f'const extraHeaders = {{ "fly-force-instance-id" : "{instance_id}" }};').encode(),
            )
            match = re.search(b'const query = ({.*?})', content)
            if match:
                new_js_object = match.group(1).decode().rstrip('}') + ", 'fly_instance_id' : '" + instance_id + "'}"
                content = content.replace(match.group(0), f'const query = {new_js_object}'.encode())
            compressed = gzip.compress(content)
            total_size = len(compressed)
            response.headers.update({'content-encoding': 'gzip', 'content-length': str(total_size)})
            result = {
                'status_code': response.status_code,
                'headers': response.headers.multi_items(),
                'content': compressed,
            }

            # NOTE: chunk large responses to stay within the SocketIO limit
            if total_size > 1_000_000:
                async def chunk_iterator() -> AsyncIterator[bytes]:
                    chunk_size = 512 * 1024
                    for i in range(0, total_size, chunk_size):
                        yield compressed[i:i + chunk_size]
                stream_id = str(uuid4())
                self.streams[stream_id] = Stream(data=chunk_iterator(), response=response)
                result['stream_id'] = stream_id
                result['total_size'] = total_size
                del result['content']

            return result

        @self.relay.on('range-request')
        async def _handle_range_request(data: dict[str, Any]) -> dict[str, Any]:
            headers: dict[str, Any] = data['headers']
            url = next(iter(u for u in core.app.urls if self.remote_url != u)) + data['path']
            data['params']['nicegui_chunk_size'] = 1024
            request = self.client.build_request(
                data['method'],
                url,
                params=data['params'],
                headers=headers,
            )
            response = await self.streaming_client.send(request, stream=True)
            stream_id = str(uuid4())
            self.streams[stream_id] = Stream(data=response.aiter_bytes(), response=response)
            return {
                'status_code': response.status_code,
                'headers': response.headers.multi_items(),
                'stream_id': stream_id,
            }

        @self.relay.on('read-stream')
        async def _handle_read_stream(stream_id: str) -> bytes | None:
            try:
                return await self.streams[stream_id].data.__anext__()
            except StopAsyncIteration:
                await _handle_close_stream(stream_id)
                return None
            except Exception:
                await _handle_close_stream(stream_id)
                raise

        @self.relay.on('close-stream')
        async def _handle_close_stream(stream_id: str) -> None:
            await self.streams[stream_id].response.aclose()
            del self.streams[stream_id]

        @self.relay.on('ready')
        def _handle_ready(data: dict[str, Any]) -> None:
            core.app.urls.add(data['device_url'])
            self.remote_url = data['device_url']
            if core.app.config.show_welcome_message:
                print(f'NiceGUI is on air at {data["device_url"]}', flush=True)

        @self.relay.on('error')
        def _handleerror(data: dict[str, Any]) -> None:
            print('Error:', data['message'], flush=True)

        @self.relay.on('handshake')
        def _handle_handshake(data: dict[str, Any]) -> bool:
            if client := Client.instances.get(data['client_id']):
                client.environ = data['environ']
                if data.get('old_tab_id'):
                    core.app.storage.copy_tab(data['old_tab_id'], data['tab_id'])
                client.tab_id = data['tab_id']
                client.on_air = True
                client.handle_handshake(data['sid'], data['document_id'], data.get('next_message_id'))
                return True
            return False

        @self.relay.on('client_disconnect')
        def _handle_client_disconnect(data: dict[str, Any]) -> None:
            self.log.debug('client disconnected.')
            if client := Client.instances.get(data['client_id']):
                client.handle_disconnect(data['sid'])

        @self.relay.on('connect')
        async def _handle_connect() -> None:
            self.log.debug('connected.')
            # NOTE: reset the warning so it can be shown again if connection breaks in the future
            helpers._shown_warnings.discard(self._host_unreachable_warning)  # pylint: disable=protected-access

        @self.relay.on('disconnect')
        async def _handle_disconnect() -> None:
            self.log.debug('disconnected.')

        @self.relay.on('connect_error')
        async def _handle_connect_error(data) -> None:
            message = data.get('message', 'Unknown error') if isinstance(data, dict) else data
            if message == 'Connection error':
                helpers.warn_once(self._host_unreachable_warning)
            else:
                self.log.warning(f'Connection error: {message}')
            await self.relay.disconnect()

        @self.relay.on('event')
        def _handle_event(data: dict[str, Any]) -> None:
            if client := Client.instances.get(data['client_id']):
                args = data['msg']['args']
                if args and isinstance(args[0], str) and args[0].startswith('{"socket_id":'):
                    arg0 = json.loads(args[0])
                    arg0['socket_id'] = client.id  # HACK: translate socket_id of ui.scene's init event
                    args[0] = json.dumps(arg0)
                client.handle_event(data['msg'])

        @self.relay.on('log')
        def _handle_log(data: dict[str, Any]) -> None:
            if client := Client.instances.get(data['client_id']):
                client.handle_log_message(data['msg'])

        @self.relay.on('javascript_response')
        def _handle_javascript_response(data: dict[str, Any]) -> None:
            if client := Client.instances.get(data['client_id']):
                client.handle_javascript_response(data['msg'])

        @self.relay.on('ack')
        def _handle_ack(data: dict[str, Any]) -> None:
            if client := Client.instances.get(data['client_id']):
                client.outbox.prune_history(data['msg']['next_message_id'])

        @self.relay.on('out_of_time')
        async def _handle_out_of_time() -> None:
            print('Sorry, you have reached the time limit of this NiceGUI On Air preview.', flush=True)
            await self.connect()

    async def connect(self) -> None:
        """Connect to the NiceGUI On Air server."""
        if self.connecting:
            self.log.debug('Already connecting.')
            return
        if self.relay.connected:
            return
        self.log.debug('Going to connect...')
        self.connecting = True
        try:
            if self.relay.connected:
                await asyncio.wait_for(self.disconnect(), timeout=5)
            self.log.debug('Connecting...')
            await self.relay.connect(
                f'{RELAY_HOST}?device_token={self.token}',
                socketio_path='/on_air/socket.io',
                transports=['websocket', 'polling'],  # favor websocket over polling
                wait_timeout=5,
            )
            assert self.relay.connected
            return
        except socketio.exceptions.ConnectionError:
            self.log.debug('Connection error.', stack_info=True)
        except ValueError:  # NOTE this sometimes happens when the internal socketio client is not yet ready
            self.log.debug('ValueError while connecting.', stack_info=True)
        except Exception:
            log.exception('Could not connect to NiceGUI On Air server.')
        finally:
            self.connecting = False

    async def disconnect(self) -> None:
        """Disconnect from the NiceGUI On Air server."""
        self.log.debug('Disconnecting...')
        if self.relay.connected:
            await self.relay.disconnect()
        for stream in self.streams.values():
            await stream.response.aclose()
        self.streams.clear()
        self.log.debug('Disconnected.')

    async def emit(self, message_type: str, data: dict[str, Any], room: str) -> None:
        """Emit a message to the NiceGUI On Air server."""
        if self.relay.connected:
            await self.relay.emit('forward', {'event': message_type, 'data': data, 'room': room})

    @staticmethod
    def is_air_target(target_id: str) -> bool:
        """Whether the given target ID is an On Air client or a SocketIO room."""
        if target_id in Client.instances:
            return Client.instances[target_id].on_air
        return target_id in core.sio.manager.rooms


def connect() -> None:
    """Connect to the NiceGUI On Air server if there is an air instance."""
    if core.air:
        background_tasks.create(core.air.connect(), name='On Air connect')


def disconnect() -> None:
    """Disconnect from the NiceGUI On Air server if there is an air instance."""
    if core.air:
        background_tasks.create(core.air.disconnect(), name='On Air disconnect')



from collections.abc import Callable
from typing import Any

from ..client import Client
from ..element import Element
from .mixins.text_element import TextElement


class Link(TextElement, component='link.js', default_classes='nicegui-link'):

    def __init__(self,
                 text: str = '',
                 target: Callable[..., Any] | str | Element = '#',
                 new_tab: bool = False,
                 ) -> None:
        """Link

        Create a hyperlink.

        To jump to a specific location within a page you can place linkable anchors with `ui.link_target("name")`
        and link to it with `ui.link(target="#name")`.

        :param text: display text
        :param target: page function, NiceGUI element on the same page or string that is a an absolute URL or relative path from base URL
        :param new_tab: open link in new tab (default: False)
        """
        super().__init__(text=text)
        if isinstance(target, str):
            self._props['href'] = target
        elif isinstance(target, Element):
            self._props['href'] = f'#{target.html_id}'
        elif callable(target):
            self._props['href'] = Client.page_routes[target]
        self._props['target'] = '_blank' if new_tab else '_self'


class LinkTarget(Element):

    def __init__(self, name: str) -> None:
        """Link target

        Create an anchor tag that can be used as inner-page target for links.

        :param name: target name
        """
        super().__init__('a')
        self._props['name'] = name



from __future__ import annotations

import io
from contextlib import suppress

from .. import json, optional_features
from ..logging import log
from .javascript import run_javascript

with suppress(ImportError):
    from PIL import Image as PIL_Image
    optional_features.register('pillow')


async def read() -> str | None:
    """Read text from the clipboard.

    Note: This function only works in secure contexts (HTTPS or localhost).
    """
    result = await run_javascript('''
        if (navigator.clipboard) {
            return navigator.clipboard.readText()
        }
        else {
            console.error('Clipboard API is only available in secure contexts (HTTPS or localhost).')
        }
    ''')
    if result is None:
        log.warning('Clipboard API is only available in secure contexts (HTTPS or localhost).')
    return result


def write(text: str) -> None:
    """Write text to the clipboard.

    Note: This function only works in secure contexts (HTTPS or localhost).

    :param text: text to write
    """
    run_javascript(f'''
        if (navigator.clipboard) {{
            navigator.clipboard.writeText({json.dumps(text)})
        }}
        else {{
            console.error('Clipboard API is only available in secure contexts (HTTPS or localhost).')
        }}
    ''')


async def read_image() -> PIL_Image.Image | None:
    """Read PIL images from the clipboard.

    Note: This function only works in secure contexts (HTTPS or localhost) and requires Pillow to be installed.

    *Added in version 2.10.0*
    """
    if not optional_features.has('pillow'):
        log.warning('Pillow is not installed, so we cannot read images from the clipboard.')
        return None
    content = await run_javascript('''
        if (navigator.clipboard) {
            const items = await navigator.clipboard.read();
            for (const item of items) {
                if (item.types.length > 0 && /^image/.test(item.types[0])) {
                    return await item.getType(item.types[0]);
                }
            }
        }
        else {
            console.error('Clipboard API is only available in secure contexts (HTTPS or localhost).');
        }
    ''', timeout=5)
    if not content:
        return None
    buffer = io.BytesIO(content)
    return PIL_Image.open(buffer)



from .mixins.text_element import TextElement


class Label(TextElement):

    def __init__(self, text: str = '') -> None:
        """Label

        Displays some text.

        :param text: the content of the label
        """
        super().__init__(tag='div', text=text)



from __future__ import annotations

import _thread
import multiprocessing as mp
import queue
import socket
import sys
import time
import warnings
from collections.abc import Callable
from contextlib import suppress
from multiprocessing.connection import Connection
from multiprocessing.synchronize import Event as MultiprocessingEvent
from threading import Event, Thread
from typing import Any

from .. import core, helpers, optional_features
from ..logging import log
from ..server import Server
from . import native
from .event_manager import event_manager

with suppress(ImportError):
    with warnings.catch_warnings():
        # webview depends on bottle which uses the deprecated CGI function (https://github.com/bottlepy/bottle/issues/1403)
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        import webview
        import webview.dom
    optional_features.register('webview')


def _open_window(
    protocol: str, host: str, port: int, title: str, width: int, height: int, fullscreen: bool, frameless: bool,
    method_queue: mp.Queue, response_queue: mp.Queue, event_sender: Connection,
) -> None:
    while not helpers.is_port_open(host, port):
        time.sleep(0.1)

    window_kwargs = {
        'url': f'{protocol}://{host}:{port}',
        'title': title,
        'width': width,
        'height': height,
        'fullscreen': fullscreen,
        'frameless': frameless,
        **core.app.native.window_args,
    }
    webview.settings.update(**core.app.native.settings)
    window = webview.create_window(**window_kwargs)
    assert window is not None
    closed = Event()
    window.events.closed += closed.set
    _bind_pywebview_events(window, event_sender)
    _start_window_method_executor(window, method_queue, response_queue, closed)
    webview.start(**core.app.native.start_args)


def _bind_pywebview_events(window: webview.Window, event_sender: Connection) -> None:
    def send(event_type: str, **kwargs: Any) -> None:
        try:
            event_sender.send({'type': event_type, 'args': kwargs})
        except OSError:
            pass

    def bind_drop() -> None:
        window.dom.document.events.dragover += \
            webview.dom.DOMEventHandler(lambda _: 0, True, False)  # type: ignore[arg-type]
        window.dom.document.events.drop += \
            webview.dom.DOMEventHandler(lambda e: send('drop', files=[  # type: ignore[arg-type]
                file_.get('pywebviewFullPath', '') for file_ in e.get('dataTransfer', {}).get('files', [])
            ]), True, False)

    window.events.shown += lambda: send('shown')
    window.events.loaded += lambda: send('loaded')
    window.events.loaded += bind_drop
    window.events.minimized += lambda: send('minimized')
    window.events.maximized += lambda: send('maximized')
    window.events.restored += lambda: send('restored')
    window.events.resized += lambda width, height: send('resized', width=width, height=height)
    window.events.moved += lambda x, y: send('moved', x=x, y=y)
    window.events.closed += lambda: send('closed')
    # NOTE: 'closing' is not bridged yet — it requires a synchronous round-trip to support vetoing the close


def _start_window_method_executor(window: webview.Window,
                                  method_queue: mp.Queue,
                                  response_queue: mp.Queue,
                                  closed: Event) -> None:
    def execute(method: Callable, args: tuple[Any, ...], kwargs: dict[str, Any]) -> None:
        try:
            response = method(*args, **kwargs)
            if response is not None or 'dialog' in method.__name__:
                response_queue.put(response)
        except Exception:
            log.exception(f'error in window.{method.__name__}')

    def window_method_executor() -> None:
        pending_executions: list[Thread] = []
        while not closed.is_set():
            try:
                method_name, args, kwargs = method_queue.get(block=False)
                if method_name == 'signal_server_shutdown':
                    if pending_executions:
                        log.warning('shutdown is possibly blocked by opened dialogs like a file picker')
                        while pending_executions:
                            pending_executions.pop().join()
                elif method_name == 'get_always_on_top':
                    response_queue.put(window.on_top)
                elif method_name == 'set_always_on_top':
                    window.on_top = args[0]
                elif method_name == 'get_position':
                    response_queue.put((int(window.x), int(window.y)))
                elif method_name == 'get_size':
                    response_queue.put((int(window.width), int(window.height)))
                else:
                    method = getattr(window, method_name)
                    if callable(method):
                        pending_executions.append(Thread(target=execute, args=(method, args, kwargs)))
                        pending_executions[-1].start()
                    else:
                        log.error(f'window.{method_name} is not callable')
            except queue.Empty:
                time.sleep(0.016)  # NOTE: avoid issue https://github.com/zauberzeug/nicegui/issues/2482 on Windows
            except Exception:
                log.exception(f'error in window.{method_name}')

    Thread(target=window_method_executor).start()


def activate(protocol: str, host: str, port: int, title: str, width: int, height: int, fullscreen: bool, frameless: bool,
             shutdown_event: MultiprocessingEvent | None = None) -> None:
    """Activate native mode."""
    def check_shutdown() -> None:
        while process.is_alive():
            time.sleep(0.1)
        if shutdown_event is not None:
            shutdown_event.set()
        Server.instance.should_exit = True
        while not core.app.is_stopped:
            time.sleep(0.1)
        _thread.interrupt_main()
        event_manager.stop()
        native.remove_queues()

    if not optional_features.has('webview'):
        log.error('Native mode is not supported in this configuration.\n'
                  'Please run "pip install pywebview" to use it.')
        sys.exit(1)

    mp.freeze_support()
    native.create_queues()
    event_manager.start()
    args = (protocol, host, port, title, width, height, fullscreen, frameless,
            native.method_queue, native.response_queue, native.event_sender)
    process = mp.Process(target=_open_window, args=args, daemon=True)
    process.start()

    Thread(target=check_shutdown, daemon=True).start()


def find_open_port(start_port: int = 8000, end_port: int = 8999) -> int:
    """Reliably find an open port in a given range.

    This function will actually try to open the port to ensure no firewall blocks it.
    This is better than, e.g., passing port=0 to uvicorn.
    """
    for port in range(start_port, end_port + 1):
        with suppress(OSError):
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind(('localhost', port))
                return port
    raise OSError('No open port found')



from typing import Any

from ..events import Handler, ValueChangeEventArguments
from .mixins.value_element import ValueElement


class ChoiceElement(ValueElement):

    def __init__(self, *,
                 tag: str | None = None,
                 options: list | dict,
                 value: Any,
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        self.options = options
        self._values: list[str] = []
        self._labels: list[str] = []
        self._update_values_and_labels()
        if not isinstance(value, list) and value is not None and value not in self._values:
            raise ValueError(f'Invalid value: {value}')
        super().__init__(tag=tag, value=value, on_value_change=on_change)
        self._update_options()

    def _update_values_and_labels(self) -> None:
        self._values = self.options if isinstance(self.options, list) else list(self.options.keys())
        self._labels = self.options if isinstance(self.options, list) else list(self.options.values())

    def _update_options(self) -> None:
        before_value = self.value
        self._props['options'] = [{'value': index, 'label': option} for index, option in enumerate(self._labels)]
        self._props[self.VALUE_PROP] = self._value_to_model_value(before_value)
        if not isinstance(before_value, list):  # NOTE: no need to update value in case of multi-select
            self.value = before_value if before_value in self._values else None

    def update(self) -> None:
        with self._props.suspend_updates():
            self._update_values_and_labels()
            self._update_options()
        super().update()

    def set_options(self, options: list | dict, *, value: Any = ...) -> None:
        """Set the options of this choice element.

        :param options: The new options.
        :param value: The new value. If not given, the current value is kept.
        """
        self.options = options
        if value is not ...:
            self.value = value
        self.update()



from __future__ import annotations

import base64
import io
import urllib.parse
from pathlib import Path
from typing import TYPE_CHECKING

from fastapi.responses import FileResponse, Response, StreamingResponse

from . import core
from .helpers import is_file
from .version import __version__

if TYPE_CHECKING:
    from .page import page


def create_favicon_route(path: str, favicon: str | Path | None) -> None:
    """Create a favicon route for the given path."""
    if is_file(favicon):
        core.app.add_route('/favicon.ico' if path == '/' else f'{path}/favicon.ico',
                           lambda _: FileResponse(favicon))  # type: ignore


def get_favicon_url(page: page, prefix: str) -> str:
    """Return the URL of the favicon for a given page."""
    favicon = page.favicon or core.app.config.favicon
    if not favicon:
        return f'{prefix}/_nicegui/{__version__}/static/favicon.ico'

    favicon = str(favicon).strip()
    if _is_remote_url(favicon):
        return favicon
    if _is_data_url(favicon):
        return favicon
    if _is_svg(favicon):
        return _svg_to_data_url(favicon)
    if _is_char(favicon):
        return _svg_to_data_url(_char_to_svg(favicon))
    if page.path == '/' or page.favicon is None:
        return f'{prefix}/favicon.ico'

    return f'{prefix}{page.path}/favicon.ico'


def get_favicon_response() -> Response:
    """Return the FastAPI response for the global favicon."""
    if not core.app.config.favicon:
        raise ValueError(f'invalid favicon: {core.app.config.favicon}')
    favicon = str(core.app.config.favicon).strip()

    if _is_svg(favicon):
        return Response(favicon, media_type='image/svg+xml')
    if _is_data_url(favicon):
        media_type, bytes_ = _data_url_to_bytes(favicon)
        return StreamingResponse(io.BytesIO(bytes_), media_type=media_type)
    if _is_char(favicon):
        return Response(_char_to_svg(favicon), media_type='image/svg+xml')

    raise ValueError(f'invalid favicon: {favicon}')


def _is_remote_url(favicon: str) -> bool:
    return favicon.startswith(('http://', 'https://'))


def _is_char(favicon: str) -> bool:
    return len(favicon) == 1 or '\ufe0f' in favicon


def _is_svg(favicon: str) -> bool:
    return favicon.strip().startswith('<svg')


def _is_data_url(favicon: str) -> bool:
    return favicon.startswith('data:')


def _char_to_svg(char: str) -> str:
    return f'''
        <svg viewBox="0 0 128 128" width="128" height="128" xmlns="http://www.w3.org/2000/svg" >
            <style>
                @supports (-moz-appearance:none) {{
                    text {{
                        font-size: 100px;
                        transform: translateY(0.1em);
                    }}
                }}
                text {{
                    font-family: Arial, sans-serif;
                }}
            </style>
            <text y=".9em" font-size="128" font-family="Georgia, sans-serif">{char}</text>
        </svg>
    '''


def _svg_to_data_url(svg: str) -> str:
    svg_urlencoded = urllib.parse.quote(svg)
    return f'data:image/svg+xml,{svg_urlencoded}'


def _data_url_to_bytes(data_url: str) -> tuple[str, bytes]:
    media_type, base64_image = data_url.split(',', 1)
    media_type = media_type.split(':')[1].split(';')[0]
    return media_type, base64.b64decode(base64_image)



from typing_extensions import Self

from ...awaitable_response import AwaitableResponse
from ...defaults import DEFAULT_PROP, resolve_defaults
from ...element import Element
from ...events import (
    GenericEventArguments,
    Handler,
    JsonEditorChangeEventArguments,
    JsonEditorSelectEventArguments,
    handle_event,
)


class JsonEditor(Element, component='json_editor.js', esm={'nicegui-json-editor': 'dist'}):

    @resolve_defaults
    def __init__(self,
                 properties: dict, *,
                 on_select: Handler[JsonEditorSelectEventArguments] | None = None,
                 on_change: Handler[JsonEditorChangeEventArguments] | None = None,
                 schema: dict | None = DEFAULT_PROP | None,
                 ) -> None:
        """JSONEditor

        An element to create a JSON editor using `JSONEditor <https://github.com/josdejong/svelte-jsoneditor>`_.
        Updates can be pushed to the editor by changing the `properties` property.

        :param properties: dictionary of JSONEditor properties
        :param on_select: callback which is invoked when some of the content has been selected
        :param on_change: callback which is invoked when the content has changed
        :param schema: optional `JSON schema <https://json-schema.org/>`_ for validating the data being edited (*added in version 2.8.0*)
        """
        super().__init__()
        self._props['properties'] = properties
        self._update_method = 'update_editor'
        self._props.set_bool('schema', schema)

        if on_select:
            self.on_select(on_select)

        if on_change:
            self.on_change(on_change)

    def on_change(self, callback: Handler[JsonEditorChangeEventArguments]) -> Self:
        """Add a callback to be invoked when the content changes."""
        def handle_on_change(e: GenericEventArguments) -> None:
            handle_event(callback, JsonEditorChangeEventArguments(sender=self, client=self.client, **e.args))
        self.on('content_change', handle_on_change, ['content', 'errors'])
        return self

    def on_select(self, callback: Handler[JsonEditorSelectEventArguments]) -> Self:
        """Add a callback to be invoked when some of the content has been selected."""
        def handle_on_select(e: GenericEventArguments) -> None:
            handle_event(callback, JsonEditorSelectEventArguments(sender=self, client=self.client, **e.args))
        self.on('content_select', handle_on_select, ['selection'])
        return self

    @property
    def properties(self) -> dict:
        """The property dictionary."""
        return self._props['properties']

    def run_editor_method(self, name: str, *args, timeout: float = 1) -> AwaitableResponse:
        """Run a method of the JSONEditor instance.

        See the `JSONEditor README <https://github.com/josdejong/svelte-jsoneditor/>`_ for a list of methods.

        If the function is awaited, the result of the method call is returned.
        Otherwise, the method is executed without waiting for a response.

        :param name: name of the method (a prefix ":" indicates that the arguments are JavaScript expressions)
        :param args: arguments to pass to the method (Python objects or JavaScript expressions)
        :param timeout: timeout in seconds (default: 1 second)

        :return: AwaitableResponse that can be awaited to get the result of the method call
        """
        return self.run_method('run_editor_method', name, *args, timeout=timeout)



from ..defaults import DEFAULT_PROP, DEFAULT_PROPS, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .label import Label
from .mixins.color_elements import TextColorElement
from .mixins.disableable_element import DisableableElement
from .mixins.value_element import ValueElement


class Knob(ValueElement, DisableableElement, TextColorElement):

    @resolve_defaults
    def __init__(self,
                 value: float = DEFAULT_PROPS['model-value'] | 0.0,
                 *,
                 min: float = DEFAULT_PROP | 0.0,  # pylint: disable=redefined-builtin
                 max: float = DEFAULT_PROP | 1.0,  # pylint: disable=redefined-builtin
                 step: float = DEFAULT_PROP | 0.01,
                 color: str | None = DEFAULT_PROP | 'primary',
                 center_color: str | None = DEFAULT_PROP | None,
                 track_color: str | None = DEFAULT_PROP | None,
                 size: str | None = DEFAULT_PROP | None,
                 show_value: bool = False,
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        """Knob

        This element is based on Quasar's `QKnob <https://quasar.dev/vue-components/knob>`_ component.
        The element is used to take a number input from the user through mouse/touch panning.

        :param value: the initial value (default: 0.0)
        :param min: the minimum value (default: 0.0)
        :param max: the maximum value (default: 1.0)
        :param step: the step size (default: 0.01)
        :param color: knob color (either a Quasar, Tailwind, or CSS color or `None`, default: "primary")
        :param center_color: color name for the center part of the component, examples: primary, teal-10
        :param track_color: color name for the track of the component, examples: primary, teal-10
        :param size: size in CSS units, including unit name or standard size name (xs|sm|md|lg|xl), examples: 16px, 2rem
        :param show_value: whether to show the value as text
        :param on_change: callback to execute when the value changes
        """
        super().__init__(tag='q-knob', value=value, on_value_change=on_change, throttle=0.05, text_color=color)

        self._props['min'] = min
        self._props['max'] = max
        self._props['step'] = step
        self._props['show-value'] = True  # NOTE: enable default slot, e.g. for nested icon
        self._props.set_optional('center-color', center_color)
        self._props.set_optional('track-color', track_color)
        self._props.set_optional('size', size)

        self.label: Label | None = None
        if show_value:
            with self:
                self.label = Label().bind_text_from(self, 'value')



"""
Custom json module. Provides dumps and loads implementations
wrapping the orjson package. If the orjson package is not available,
the standard Python json module is used.

This custom module is required in order to override the json-module used
in socketio.AsyncServer, which expects a module as parameter
to override Python's default json module.
"""

try:
    from .orjson_wrapper import NiceGUIJSONResponse, dumps, loads
except ImportError:
    from .builtin_wrapper import NiceGUIJSONResponse, dumps, loads  # type: ignore


__all__ = [
    'NiceGUIJSONResponse',
    'dumps',
    'loads',
]



from ..defaults import DEFAULT_PROP, resolve_defaults
from ..events import Handler, ValueChangeEventArguments
from .mixins.value_element import ValueElement


class DarkMode(ValueElement, component='dark_mode.js'):
    VALUE_PROP = 'value'

    @resolve_defaults
    def __init__(self,
                 value: bool | None = DEFAULT_PROP | False, *,
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        """Dark mode

        You can use this element to enable, disable or toggle dark mode on the page.
        The value `None` represents auto mode, which uses the client's system preference.

        Note that this element overrides the `dark` parameter of the `ui.run` function and page decorators.

        :param value: Whether dark mode is enabled. If None, dark mode is set to auto.
        :param on_change: Callback that is invoked when the value changes.
        """
        super().__init__(value=value, on_value_change=on_change)

    def enable(self) -> None:
        """Enable dark mode."""
        self.value = True

    def disable(self) -> None:
        """Disable dark mode."""
        self.value = False

    def toggle(self) -> None:
        """Toggle dark mode.

        This method will raise a ValueError if dark mode is set to auto.
        """
        if self.value is None:
            raise ValueError('Cannot toggle dark mode when it is set to auto.')
        self.value = not self.value

    def auto(self) -> None:
        """Set dark mode to auto.

        This will use the client's system preference.
        """
        self.value = None



from __future__ import annotations

import inspect
import re
import weakref
from collections.abc import Callable, Iterator, Sequence
from copy import copy
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, cast

from typing_extensions import Self

from . import core, events, helpers, json, storage
from .awaitable_response import AwaitableResponse, NullResponse
from .classes import Classes
from .context import context
from .dependencies import (
    Component,
    Library,
    register_dynamic_resource,
    register_esm,
    register_library,
    register_resource,
    register_vue_component,
)
from .elements.mixins.visibility import Visibility
from .event_listener import EventListener
from .props import Props
from .slot import Slot
from .style import Style
from .version import __version__

if TYPE_CHECKING:
    from .client import Client

# https://www.w3.org/TR/xml/#sec-common-syn
TAG_START_CHAR = r':|[A-Z]|_|[a-z]|[\u00C0-\u00D6]|[\u00D8-\u00F6]|[\u00F8-\u02FF]|[\u0370-\u037D]|[\u037F-\u1FFF]|[\u200C-\u200D]|[\u2070-\u218F]|[\u2C00-\u2FEF]|[\u3001-\uD7FF]|[\uF900-\uFDCF]|[\uFDF0-\uFFFD]|[\U00010000-\U000EFFFF]'
TAG_CHAR = TAG_START_CHAR + r'|-|\.|[0-9]|\u00B7|[\u0300-\u036F]|[\u203F-\u2040]'
TAG_PATTERN = re.compile(fr'^({TAG_START_CHAR})({TAG_CHAR})*$')


class Element(Visibility):
    component: Component | None = None
    exposed_libraries: ClassVar[list[Library]] = []
    _default_props: ClassVar[dict[str, Any]] = {}
    _default_classes: ClassVar[list[str]] = []
    _default_style: ClassVar[dict[str, str]] = {}

    def __init__(self, tag: str | None = None, *, _client: Client | None = None) -> None:
        """Generic Element

        This class is the base class for all other UI elements.
        But you can use it to create elements with arbitrary HTML tags.

        :param tag: HTML tag of the element
        :param _client: client for this element (for internal use only)
        """
        super().__init__()
        client = _client or context.client
        self._client = weakref.ref(client)
        self.id = client.next_element_id
        client.next_element_id += 1
        self.tag = tag if tag else self.component.tag if self.component else 'div'
        if not TAG_PATTERN.match(self.tag):
            raise ValueError(f'Invalid HTML tag: {self.tag}')
        self._classes: Classes[Self] = Classes(self._default_classes, element=cast(Self, self))
        self._style: Style[Self] = Style(self._default_style, element=cast(Self, self))
        self._props: Props[Self] = Props(self._default_props, element=cast(Self, self))
        self._markers: list[str] = []
        self._event_listeners: dict[str, EventListener] = {}
        self._text: str | None = None
        self.slots: dict[str, Slot] = {}
        self.default_slot = self.add_slot('default')
        self._update_method: str | None = None
        self._deleted: bool = False

        client.elements[self.id] = self
        self._parent_slot: weakref.ref[Slot] | None = None
        slot_stack = context.slot_stack
        if slot_stack:
            parent_slot = slot_stack[-1]
            parent_slot.children.append(self)
            self._parent_slot = weakref.ref(parent_slot)

        client.outbox.enqueue_update(self)
        if self._parent_slot:
            client.outbox.enqueue_update(parent_slot.parent)

        self._props.add_rename('resource_path', 'resource-path')  # DEPRECATED: remove in NiceGUI 4.0
        self._props.add_rename('dynamic_resource_path', 'dynamic-resource-path')  # DEPRECATED: remove in NiceGUI 4.0

    def __init_subclass__(cls, *,
                          component: str | Path | None = None,
                          dependencies: list[str | Path] = [],  # noqa: B006
                          esm: dict[str, str] | None = None,
                          default_classes: str | None = None,
                          default_style: str | None = None,
                          default_props: str | None = None,
                          ) -> None:
        super().__init_subclass__()
        base = Path(inspect.getfile(cls)).parent

        def glob_absolute_paths(file: str | Path) -> list[Path]:
            path = Path(file)
            if not path.is_absolute():
                path = base / path
            return sorted(path.parent.glob(path.name), key=lambda p: p.stem)

        cls.component = copy(cls.component)
        cls.exposed_libraries = copy(cls.exposed_libraries)
        if component:
            max_time = max((path.stat().st_mtime for path in glob_absolute_paths(component)), default=None)
            for path in glob_absolute_paths(component):
                cls.component = register_vue_component(path, max_time=max_time)
        for library in dependencies:
            max_time = max((path.stat().st_mtime for path in glob_absolute_paths(library)), default=None)
            for path in glob_absolute_paths(library):
                cls.exposed_libraries.append(register_library(path, max_time=max_time))
        for key, esm_path in (esm or {}).items():
            path = Path(esm_path)
            if not path.is_absolute():
                path = base / path
            max_time = max((path.stat().st_mtime for path in glob_absolute_paths(path)), default=None)
            register_esm(key, path, max_time=max_time)

        cls._default_props = copy(cls._default_props)
        cls._default_classes = copy(cls._default_classes)
        cls._default_style = copy(cls._default_style)
        cls.default_classes(default_classes)
        cls.default_style(default_style)
        cls.default_props(default_props)

    @property
    def client(self) -> Client:
        """The client this element belongs to."""
        client = self._client()
        if client is None:
            raise RuntimeError('The client this element belongs to has been deleted.')
        return client

    @property
    def parent_slot(self) -> Slot | None:
        """The parent slot of the element."""
        if self._parent_slot is None:
            return None
        parent_slot = self._parent_slot()
        if parent_slot is None:
            raise RuntimeError('The parent slot of the element has been deleted.')
        return parent_slot

    @parent_slot.setter
    def parent_slot(self, value: Slot | None) -> None:
        self._parent_slot = weakref.ref(value) if value else None

    def add_resource(self, path: str | Path) -> None:
        """Add a resource to the element.

        :param path: path to the resource (e.g. folder with CSS and JavaScript files)
        """
        path_ = Path(path)
        resource = register_resource(path_, max_time=path_.stat().st_mtime)
        self._props['resource-path'] = f'/_nicegui/{__version__}/resources/{resource.key}'

    def add_dynamic_resource(self, name: str, function: Callable) -> None:
        """Add a dynamic resource to the element which returns the result of a function.

        :param name: name of the resource
        :param function: function that returns the resource response
        """
        register_dynamic_resource(name, function)
        self._props['dynamic-resource-path'] = f'/_nicegui/{__version__}/dynamic_resources'

    def add_slot(self, name: str, template: str | None = None) -> Slot:
        """Add a slot to the element.

        NiceGUI is using the slot concept from Vue:
        Elements can have multiple slots, each possibly with a number of children.
        Most elements only have one slot, e.g. a `ui.card` (QCard) only has a default slot.
        But more complex elements like `ui.table` (QTable) can have more slots like "header", "body" and so on.
        If you nest NiceGUI elements via with `ui.row(): ...` you place new elements inside of the row's default slot.
        But if you use with `table.add_slot(...): ...`, you enter a different slot.

        The slot stack helps NiceGUI to keep track of which slot is currently used for new elements.
        The `parent` field holds a reference to its element.
        Whenever an element is entered via a `with` expression, its default slot is automatically entered as well.

        :param name: name of the slot
        :param template: Vue template of the slot
        :return: the slot
        """
        self.slots[name] = Slot(self, name, template)
        return self.slots[name]

    def __enter__(self) -> Self:
        self.default_slot.__enter__()
        return self

    def __exit__(self, *_) -> None:
        self.default_slot.__exit__(*_)

    def __iter__(self) -> Iterator[Element]:
        for slot in self.slots.values():
            yield from slot

    def _collect_slot_dict(self) -> dict[str, Any]:
        return {
            name: {
                'ids': [child.id for child in slot],
                **({'template': slot.template} if slot.template is not None else {}),
            }
            for name, slot in self.slots.items()
            if slot != self.default_slot
        }

    def _to_dict(self) -> dict[str, Any]:
        return {
            'tag': self.tag,
            **({'text': self._text} if self._text is not None else {}),
            **{
                key: value
                for key, value in {
                    'class': self._classes,
                    'style': self._style,
                    'props': self._props,
                    'slots': self._collect_slot_dict(),
                    'children': [child.id for child in self.default_slot.children],
                    'events': [listener.to_dict() for listener in self._event_listeners.values()],
                    'update_method': self._update_method,
                }.items()
                if value
            },
        }

    @property
    def classes(self) -> Classes[Self]:
        """The classes of the element."""
        return self._classes

    @classmethod
    def default_classes(cls,
                        add: str | None = None, *,
                        remove: str | None = None,
                        toggle: str | None = None,
                        replace: str | None = None) -> type[Self]:
        """Apply, remove, toggle, or replace default HTML classes.

        This allows modifying the look of the element or its layout using `Tailwind <https://tailwindcss.com/>`_ or `Quasar <https://quasar.dev/>`_ classes.

        Removing or replacing classes can be helpful if predefined classes are not desired.
        All elements of this class will share these HTML classes.
        These must be defined before element instantiation.

        :param add: whitespace-delimited string of classes
        :param remove: whitespace-delimited string of classes to remove from the element
        :param toggle: whitespace-delimited string of classes to toggle (*added in version 2.7.0*)
        :param replace: whitespace-delimited string of classes to use instead of existing ones
        """
        cls._default_classes = Classes.update_list(cls._default_classes, add, remove, toggle, replace)
        return cls

    @property
    def style(self) -> Style[Self]:
        """The style of the element."""
        return self._style

    @classmethod
    def default_style(cls,
                      add: str | None = None, *,
                      remove: str | None = None,
                      replace: str | None = None) -> type[Self]:
        """Apply, remove, or replace default CSS definitions.

        Removing or replacing styles can be helpful if the predefined style is not desired.
        All elements of this class will share these CSS definitions.
        These must be defined before element instantiation.

        :param add: semicolon-separated list of styles to add to the element
        :param remove: semicolon-separated list of styles to remove from the element
        :param replace: semicolon-separated list of styles to use instead of existing ones
        """
        if replace is not None:
            cls._default_style.clear()
        for key in Style.parse(remove):
            cls._default_style.pop(key, None)
        cls._default_style.update(Style.parse(add))
        cls._default_style.update(Style.parse(replace))
        return cls

    @property
    def props(self) -> Props[Self]:
        """The props of the element."""
        return self._props

    @classmethod
    def default_props(cls,
                      add: str | None = None, *,
                      remove: str | None = None) -> type[Self]:
        """Add or remove default props.

        This allows modifying the look of the element or its layout using `Quasar <https://quasar.dev/>`_ props.
        Since props are simply applied as HTML attributes, they can be used with any HTML element.
        All elements of this class will share these props.
        These must be defined before element instantiation.

        Boolean properties are assumed ``True`` if no value is specified.

        :param add: whitespace-delimited list of either boolean values or key=value pair to add
        :param remove: whitespace-delimited list of property keys to remove
        """
        for key in Props.parse(remove):
            if key in cls._default_props:
                del cls._default_props[key]
        for key, value in Props.parse(add).items():
            cls._default_props[key] = value
        return cls

    def mark(self, *markers: str) -> Self:
        """Replace markers of the element.

        Markers are used to identify elements for querying with `ElementFilter </documentation/element_filter>`_
        which is heavily used in testing
        but can also be used to reduce the number of global variables or passing around dependencies.

        :param markers: list of strings or single string with whitespace-delimited markers; replaces existing markers
        """
        self._markers = [word for marker in markers for word in marker.split()]
        return self

    def tooltip(self, text: str) -> Self:
        """Add a tooltip to the element.

        :param text: text of the tooltip
        """
        from .elements.tooltip import Tooltip  # pylint: disable=import-outside-toplevel, cyclic-import
        Tooltip(text).props['target'] = f'#{self.html_id}'
        return self

    def on(self,
           type: str,  # pylint: disable=redefined-builtin
           handler: events.Handler[events.GenericEventArguments] | None = None,
           args: None | Sequence[str] | Sequence[Sequence[str] | None] = None,
           *,
           throttle: float = 0.0,
           leading_events: bool = True,
           trailing_events: bool = True,
           js_handler: str = '(...args) => emit(...args)',
           ) -> Self:
        """Subscribe to an event.

        The event handler can be a Python function, a JavaScript function or a combination of both:

        - If you want to handle the event on the server with all (serializable) event arguments,
          use a Python ``handler``.
        - If you want to handle the event on the client side without emitting anything to the server,
          use ``js_handler`` with a JavaScript function handling the event.
        - If you want to handle the event on the server with a subset or transformed version of the event arguments,
          use ``js_handler`` with a JavaScript function emitting the transformed arguments using ``emit()``, and
          use a Python ``handler`` to handle these arguments on the server side.
          The ``js_handler`` can also decide to selectively emit arguments to the server,
          in which case the Python ``handler`` will not always be called.

        Note that the arguments ``throttle``, ``leading_events``, and ``trailing_events`` are only relevant
        when emitting events to the server.

        *Updated in version 2.18.0: Both handlers can be specified at the same time.*

        :param type: name of the event (e.g. "click", "mousedown", or "update:model-value")
        :param handler: callback that is called upon occurrence of the event
        :param args: arguments included in the event message sent to the event handler (default: ``None`` meaning all)
        :param throttle: minimum time (in seconds) between event occurrences (default: 0.0)
        :param leading_events: whether to trigger the event handler immediately upon the first event occurrence (default: ``True``)
        :param trailing_events: whether to trigger the event handler after the last event occurrence (default: ``True``)
        :param js_handler: JavaScript function that is handling the event on the client (default: "(...args) => emit(...args)")
        """
        if handler or js_handler:
            listener = EventListener(
                element_id=self.id,
                type=helpers.event_type_to_camel_case(type),
                args=[args] if args and isinstance(args[0], str) else args,  # type: ignore
                handler=handler,
                js_handler=None if js_handler == '(...args) => emit(...args)' else js_handler,
                throttle=throttle,
                leading_events=leading_events,
                trailing_events=trailing_events,
                request=storage.request_contextvar.get(),
            )
            self._event_listeners[listener.id] = listener
            self.update()
        return self

    def _handle_event(self, msg: dict) -> None:
        listener = self._event_listeners[msg['listener_id']]
        storage.request_contextvar.set(listener.request)
        args = events.GenericEventArguments(sender=self, client=self.client, args=msg['args'])
        events.handle_event(listener.handler, args)

    def update(self) -> None:
        """Update the element on the client side."""
        if self.is_deleted:
            return
        self.client.outbox.enqueue_update(self)

    def run_method(self, name: str, *args: Any, timeout: float = 1) -> AwaitableResponse:
        """Run a method on the client side.

        If the function is awaited, the result of the method call is returned.
        Otherwise, the method is executed without waiting for a response.

        :param name: name of the method
        :param args: arguments to pass to the method
        :param timeout: maximum time to wait for a response (default: 1 second)
        """
        if not core.loop:
            return NullResponse()
        return self.client.run_javascript(
            f'return runMethod({self.id}, {json.dumps(name)}, {json.dumps(args)})', timeout=timeout,
        )

    def get_computed_prop(self, prop_name: str, *, timeout: float = 1) -> AwaitableResponse:
        """Return a computed property.

        This function should be awaited so that the computed property is properly returned.

        :param prop_name: name of the computed prop
        :param timeout: maximum time to wait for a response (default: 1 second)
        """
        if not core.loop:
            return NullResponse()
        return self.client.run_javascript(
            f'return getComputedProp({self.id}, {json.dumps(prop_name)})', timeout=timeout,
        )

    def ancestors(self, *, include_self: bool = False) -> Iterator[Element]:
        """Iterate over the ancestors of the element.

        :param include_self: whether to include the element itself in the iteration
        """
        if include_self:
            yield self
        parent_slot = self.parent_slot
        if parent_slot:
            yield from parent_slot.parent.ancestors(include_self=True)

    def descendants(self, *, include_self: bool = False) -> Iterator[Element]:
        """Iterate over the descendants of the element.

        :param include_self: whether to include the element itself in the iteration
        """
        if include_self:
            yield self
        for child in self:
            yield from child.descendants(include_self=True)

    def clear(self) -> Self:
        """Remove all child elements."""
        self.client.remove_elements(self.descendants())
        for slot in self.slots.values():
            slot.children.clear()
        self.update()
        return self

    def move(self,
             target_container: Element | None = None,
             target_index: int = -1, *,
             target_slot: str | None = None) -> None:
        """Move the element to another container.

        :param target_container: container to move the element to (default: the parent container)
        :param target_index: index within the target slot (default: append to the end)
        :param target_slot: slot within the target container (default: default slot)
        """
        parent_slot = self.parent_slot
        assert parent_slot is not None
        parent_slot.children.remove(self)
        parent_slot.parent.update()
        target_container = target_container or parent_slot.parent

        if target_slot is None:
            parent_slot = target_container.default_slot
            self.parent_slot = parent_slot
        elif target_slot in target_container.slots:
            parent_slot = target_container.slots[target_slot]
            self.parent_slot = parent_slot
        else:
            raise ValueError(f'Slot "{target_slot}" does not exist in the target container. '
                             f'Add it first using `add_slot("{target_slot}")`.')

        target_index = target_index if target_index >= 0 else len(parent_slot.children)
        parent_slot.children.insert(target_index, self)

        target_container.update()

    def remove(self, element: Element | int) -> None:
        """Remove a child element.

        :param element: either the element instance or its ID
        """
        if isinstance(element, int):
            children = list(self)
            element = children[element]
        self.client.remove_elements(element.descendants(include_self=True))
        parent_slot = element.parent_slot
        assert parent_slot is not None
        parent_slot.children.remove(element)
        self.update()

    def delete(self) -> None:
        """Delete the element and all its children."""
        parent_slot = self.parent_slot
        assert parent_slot is not None
        parent_slot.parent.remove(self)

    def _handle_delete(self) -> None:
        """Called when the element is deleted.

        This method can be overridden in subclasses to perform cleanup tasks.
        """
        for slot in self.slots.values():
            slot.children.clear()
        self._event_listeners.clear()

    @property
    def is_deleted(self) -> bool:
        """Whether the element has been deleted."""
        return self._deleted

    def __str__(self) -> str:
        result = self.tag if type(self) is Element else self.__class__.__name__  # pylint: disable=unidiomatic-typecheck

        def shorten(content: Any, length: int = 20) -> str:
            text = str(content).replace('\n', ' ').replace('\r', ' ')
            return text[:length].strip() + '...' if len(text) > length else text

        additions = []
        if self._markers:
            additions.append(f'markers={", ".join(self._markers)}')
        if self._text:
            additions.append(f'text={shorten(self._text)}')
        if hasattr(self, 'content') and self.content:  # pylint: disable=no-member
            additions.append(f'content={shorten(self.content)}')  # pylint: disable=no-member
        IGNORED_PROPS = {'loopback', 'color', 'view', 'innerHTML', 'dynamic-resource-path'}
        additions += [
            f'{key}={shorten(value)}'
            for key, value in self._props.items()
            if not key.startswith('_') and key not in IGNORED_PROPS and value
        ]
        if not self.visible:
            additions.append(f'visible={self.visible}')
        if additions:
            result += f' [{", ".join(additions)}]'

        for child in self.default_slot.children:
            for line in str(child).split('\n'):
                result += f'\n {line}'

        return result

    @property
    def html_id(self) -> str:
        """The ID of the element in the HTML DOM.

        *Added in version 2.16.0*
        """
        return f'c{self.id}'



from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal

from ..language import Language


@dataclass(kw_only=True, slots=True)
class AppConfig:
    endpoint_documentation: Literal['none', 'internal', 'page', 'all'] = 'none'
    socket_io_js_query_params: dict = field(default_factory=dict)
    socket_io_js_extra_headers: dict = field(default_factory=dict)
    socket_io_js_transports: list[Literal['websocket', 'polling']] = \
        field(default_factory=lambda: ['websocket', 'polling'])  # NOTE: we favor websocket
    quasar_config: dict = \
        field(default_factory=lambda: {
            'brand': {},
            'loadingBar': {
                'color': 'primary',
                'skipHijack': False,
            },
        })
    vue_config_script: str = r'''
        app.use(Quasar, {config: vue_config});
        applyColors(vue_config.brand);
        Quasar.lang.set(Quasar.lang[language.replace('-', '')]);
        darkSetter = (dark) => Quasar.Dark.set(dark === None ? "auto" : dark);
        setDark(dark);
    '''

    reload: bool = field(init=False)
    title: str = field(init=False)
    viewport: str = field(init=False)
    favicon: str | Path | None = field(init=False)
    dark: bool | None = field(init=False)
    language: Language = field(init=False)
    binding_refresh_interval: float | None = field(init=False)
    reconnect_timeout: float = field(init=False)
    message_history_length: int = field(init=False)
    cache_control_directives: str = field(init=False)
    tailwind: bool = field(init=False)
    unocss: Literal['mini', 'wind3', 'wind4'] | None = field(init=False)
    prod_js: bool = field(init=False)
    show_welcome_message: bool = field(init=False)
    _has_run_config: bool = False

    def add_run_config(self,
                       *,
                       reload: bool,
                       title: str,
                       viewport: str,
                       favicon: str | Path | None,
                       dark: bool | None,
                       language: Language,
                       binding_refresh_interval: float | None,
                       reconnect_timeout: float,
                       message_history_length: int,
                       cache_control_directives: str = 'public, max-age=31536000, immutable, stale-while-revalidate=31536000',
                       tailwind: bool,
                       unocss: Literal['mini', 'wind3', 'wind4'] | None,
                       prod_js: bool,
                       show_welcome_message: bool,
                       ) -> None:
        """Add the run config to the app config."""
        self.reload = reload
        self.title = title
        self.viewport = viewport
        self.favicon = favicon
        self.dark = dark
        self.language = language
        self.binding_refresh_interval = binding_refresh_interval
        self.reconnect_timeout = reconnect_timeout
        self.message_history_length = message_history_length
        self.cache_control_directives = cache_control_directives
        self.tailwind = tailwind if unocss is None else False
        self.unocss = unocss
        self.prod_js = prod_js
        self.show_welcome_message = show_welcome_message
        self._has_run_config = True

    @property
    def has_run_config(self) -> bool:
        """Return whether the run config has been added."""
        return self._has_run_config



import importlib.util
from typing import Literal

_optional_features: set[str] = set()

FEATURE = Literal[
    'altair',
    'anywidget',
    'highcharts',
    'matplotlib',
    'pandas',
    'pillow',
    'plotly',
    'polars',
    'pyecharts',
    'redis',
    'webview',
]


def register(feature: FEATURE) -> None:
    """Register an optional feature."""
    _optional_features.add(feature)


def try_register(feature: FEATURE, *, package: str | None = None) -> None:
    """Register an optional feature if the corresponding package is installed."""
    try:
        if importlib.util.find_spec(package or feature):
            _optional_features.add(feature)
    except (ModuleNotFoundError, ValueError):
        pass


def has(feature: FEATURE) -> bool:
    """Check if an optional feature is registered."""
    return feature in _optional_features



from typing import Any

from ..defaults import DEFAULT_PROPS, resolve_defaults
from ..events import GenericEventArguments, Handler, ValueChangeEventArguments
from .choice_element import ChoiceElement
from .mixins.disableable_element import DisableableElement


class Radio(ChoiceElement, DisableableElement):

    @resolve_defaults
    def __init__(self,
                 options: list | dict, *,
                 value: Any = DEFAULT_PROPS['model-value'] | None,
                 on_change: Handler[ValueChangeEventArguments] | None = None,
                 ) -> None:
        """Radio Selection

        This element is based on Quasar's `QOptionGroup <https://quasar.dev/vue-components/option-group>`_ component.

        The options can be specified as a list of values, or as a dictionary mapping values to labels.
        After manipulating the options, call `update()` to update the options in the UI.

        :param options: a list ['value1', ...] or dictionary `{'value1':'label1', ...}` specifying the options
        :param value: the initial value
        :param on_change: callback to execute when selection changes
        """
        super().__init__(tag='q-option-group', options=options, value=value, on_change=on_change)

    def _event_args_to_value(self, e: GenericEventArguments) -> Any:
        return self._values[e.args]

    def _value_to_model_value(self, value: Any) -> Any:
        return self._values.index(value) if value in self._values else None
